from sign_language_translator.vision.landmarks.connections import (
    BaseConnections,
    MediapipeConnections,
    get_connections,
)
from sign_language_translator.vision.landmarks.display import MatPlot3D
from sign_language_translator.vision.landmarks.landmarks import Landmarks

__all__ = [
    "Landmarks",
    "BaseConnections",
    "MediapipeConnections",
    "MatPlot3D",
    "get_connections",
]



"""This module defines the ConcatenativeSynthesis class, which represents a rule based model for translating text to sign language."""

from __future__ import annotations

import random
from enum import Enum
from typing import List, Type, Union

from sign_language_translator.config.enums import (
    SignEmbeddingModels,
    SignFormats,
    normalize_short_code,
)
from sign_language_translator.languages import get_sign_language, get_text_language
from sign_language_translator.languages.sign import SignLanguage
from sign_language_translator.languages.text import TextLanguage
from sign_language_translator.models.text_to_sign.t2s_model import TextToSignModel
from sign_language_translator.vision._utils import get_sign_wrapper_class
from sign_language_translator.vision.sign.sign import Sign


class ConcatenativeSynthesis(TextToSignModel):
    """A class representing a Rule-Based model for translating text to sign language
    by concatenating sign language videos.
    """

    def __init__(
        self,
        text_language: Union[str, TextLanguage, Enum],
        sign_language: Union[str, SignLanguage, Enum],
        sign_format: Union[str, Type[Sign]],
        sign_embedding_model: Union[str, Enum, None] = None,
    ) -> None:
        """
        Args:
            text_language (str | TextLanguage | Enum): (source) The text language processor object or its identifier. (e.g. "urdu" or `slt.languages.text.Urdu()`. See `slt.TextLanguageCodes` for all options.)
            sign_language (str | SignLanguage | Enum): (target) The sign language processor object or its identifier. (e.g. "pk-sl" or `slt.languages.sign.PakistanSignLanguage()`. See `slt.SignLanguageCodes` for all options.)
            sign_format (str | Type[Sign]): (format) The sign features used for mapping labels to sign features. (e.g. "video" or `slt.vision.Video` or "landmarks" or `slt.vision.Landmarks`. See `slt.SignFormatCodes` for all options.)
            sign_embedding_model (str | Enum | None, optional): The name of the model used for extracting features from the signs in available datasets. Not required for Video sign_format. (e.g. "mediapipe-world". See `slt.enums.SignEmbeddingModels` for all options.)
        """
        self._text_language = None
        self._sign_language = None
        self._sign_format = None
        self._sign_embedding_model = None

        self.text_language = text_language
        self.sign_language = sign_language
        self.sign_format = sign_format
        self.sign_embedding_model = sign_embedding_model

    @property
    def text_language(self) -> TextLanguage:
        """An object of `slt.languages.text.TextLanguage` class or its child that defines preprocessing, tokenization & other NLP functions."""
        if self._text_language is None:
            raise ValueError("Text language is not set.")
        return self._text_language

    @text_language.setter
    def text_language(self, text_language: Union[str, TextLanguage, Enum]):
        self._text_language = self.__get_text_language_object(text_language)

    @property
    def sign_language(self) -> SignLanguage:
        """An object of `slt.languages.sign.SignLanguage` class or its child that defines the mapping rules & grammar of a sign language."""
        if self._sign_language is None:
            raise ValueError("Sign language is not set.")
        return self._sign_language

    @sign_language.setter
    def sign_language(self, sign_language: Union[str, SignLanguage, Enum]):
        self._sign_language = self.__get_sign_language_object(sign_language)

    @property
    def sign_format(self) -> Type[Sign]:
        """
        The format of the sign language (e.g. `slt.Vision.sign.sign.Sign` or subclass).

        Class that wraps the sign language features e.g. raw videos or landmarks.
        This class can load the signs from available datasets and concatenate its objects.
        e.g. `slt.Video` or `slt.Landmarks` class.
        """
        if self._sign_format is None:
            raise ValueError("Sign format is not set.")
        return self._sign_format

    @sign_format.setter
    def sign_format(self, sign_format: Union[str, Type[Sign], Enum]) -> None:
        self._sign_format = self.__get_sign_format_class(sign_format)

        if self._sign_format.name() == SignFormats.VIDEO.value:
            self.sign_embedding_model = None

    @property
    def sign_embedding_model(self) -> Union[str, None]:
        """The name of the model which was used for extracting features from the signs.
        This name is used in the filenames of the preprocessed signs dataset."""
        if (
            self._sign_embedding_model is None
            and self.sign_format.name() != SignFormats.VIDEO.value
        ):
            raise ValueError(
                f"Sign embedding model is not set while using '{self.sign_format.name()}' sign format."
            )
        return self._sign_embedding_model

    @sign_embedding_model.setter
    def sign_embedding_model(self, model: Union[str, Enum, None]) -> None:
        if model is not None:
            if self.sign_format.name() == SignFormats.VIDEO.value:
                raise ValueError(
                    f"Can not set sign_embedding_model for {SignFormats.VIDEO.value} format because it uses raw video files."
                )

            normalized_model = normalize_short_code(model)
            if normalized_model not in SignEmbeddingModels:
                allowed = list(SignEmbeddingModels._value2member_map_.keys())
                raise ValueError(f"Invalid sign model. '{model}' not in {allowed}.")
            model = normalized_model

        elif self.sign_format.name() != SignFormats.VIDEO.value:
            raise ValueError(
                f"Sign embedding model is required for '{self.sign_format.name()}' sign_format."
            )

        self._sign_embedding_model = model

    def translate(self, text: str, *args, **kwargs) -> Sign:
        """
        Translate text to sign language.

        Args:
            text: The input text to be translated.

        Returns:
            The translated sign language sentence.

        """

        sign_language_sentence = None
        video_labels = []

        text = self.text_language.preprocess(text)
        sentences = self.text_language.sentence_tokenize(text)
        for sentence in sentences:
            tokens = self.text_language.tokenize(sentence)
            tags = self.text_language.get_tags(tokens)

            tokens, tags, contexts = self.sign_language.restructure_sentence(
                tokens, tags=tags
            )
            sign_dicts = self.sign_language.tokens_to_sign_dicts(
                tokens, tags=tags, contexts=contexts
            )

            video_labels.extend(
                [
                    label
                    for sign_dict in sign_dicts
                    for label in random.choices(
                        sign_dict[self.sign_language.SignDictKeys.SIGNS.value],
                        weights=sign_dict[self.sign_language.SignDictKeys.WEIGHTS.value],  # type: ignore
                        k=1,
                    )[0]
                ]
            )

        signs = self._map_labels_to_sign(video_labels)
        # TODO: Trim signs where hand is just being raised from or lowered to the resting position
        sign_language_sentence = self.sign_format.concatenate(signs)

        return sign_language_sentence

    def _map_labels_to_sign(self, labels: List[str]) -> List[Sign]:
        return [
            self.sign_format.load_asset(self._prepare_resource_name(label))
            for label in labels
        ]

    def _prepare_resource_name(self, label, person=None, camera=None, sep="_"):
        if person is not None:
            label = f"{label}{sep}{person}"
        if camera is not None:
            label = f"{label}{sep}{camera}"

        directory = self.sign_format.name().rstrip("s") + "s"

        extension = "unknown"
        if self.sign_format.name() == SignFormats.VIDEO.value:
            extension = "mp4"
        elif self.sign_format.name() == SignFormats.LANDMARKS.value:
            extension = "csv"  # todo: generalize sign filenames

        if self.sign_embedding_model is not None:
            extension = (
                f"{self.sign_format.name()}-{self.sign_embedding_model}.{extension}"
            )

        name = f"{directory}/{label}.{extension}"

        return name

    def __get_text_language_object(
        self, text_language: Union[str, TextLanguage, Enum]
    ) -> TextLanguage:
        if isinstance(text_language, (str, Enum)):
            return get_text_language(text_language)
        if isinstance(text_language, TextLanguage):
            return text_language
        raise TypeError(f"Expected str or TextLanguage, got {type(text_language) = }.")

    def __get_sign_language_object(
        self, sign_language: Union[str, SignLanguage, Enum]
    ) -> SignLanguage:
        if isinstance(sign_language, (str, Enum)):
            return get_sign_language(sign_language)
        if isinstance(sign_language, SignLanguage):
            return sign_language
        raise TypeError(f"Expected str or SignLanguage, got {type(sign_language) = }.")

    def __get_sign_format_class(
        self, sign_format: Union[str, Type[Sign], Enum]
    ) -> Type[Sign]:
        if isinstance(sign_format, (str, Enum)):
            return get_sign_wrapper_class(sign_format)
        if isinstance(sign_format, type) and issubclass(sign_format, Sign):
            return sign_format
        raise TypeError(f"Expected str or type[Sign], got {sign_format = }.")



from __future__ import annotations

__all__ = [
    "Video",
]

import struct
from copy import copy, deepcopy
from mimetypes import guess_type
from os import makedirs, remove
from os.path import abspath, dirname, isfile, join
from typing import (
    Callable,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from warnings import warn

import cv2  # TODO: Compile OpenCV with GPU
import numpy as np
import torch
from numpy.typing import NDArray
from tqdm.auto import tqdm

from sign_language_translator.config.assets import Assets
from sign_language_translator.config.enums import SignFormats
from sign_language_translator.utils import in_jupyter_notebook, validate_path_exists
from sign_language_translator.vision.sign.sign import Sign
from sign_language_translator.vision.utils import (
    _normalize_args_index_and_timestamp,
    _validate_and_normalize_slices,
)
from sign_language_translator.vision.video.display import VideoDisplay
from sign_language_translator.vision.video.video_iterators import (
    IterableFrames,
    SequenceFrames,
    VideoCaptureFrames,
    VideoFrames,
)


class Video(Sign, VideoFrames):
    """A class to represent and manipulate videos or sequence of images.
    Inherits from the `slt.vision.Sign` class.

    Args:
        sign (Union[ str, Sequence[Union[NDArray[np.uint8], torch.Tensor]], NDArray[np.uint8], torch.Tensor, Generator[NDArray[np.uint8], None, None], ]): The video source. Can be a path to a video or image file, a sequence of frames, a single frame, or a generator of frames.

    Methods:
        load(): Load a video from the specified path.
        load_asset(): Load a video asset by its label from the built-in datasets.
        name(): Returns the name of the sign format.
        numpy(): Convert the video frames to a NumPy array.
        torch(): Convert the video frames to a PyTorch tensor.
        trim(): Cut the video to a specific time range or index range.
        stack(): Stack a list of Video objects along a specified dimension.
        concatenate(): Concatenate a sequence of Video objects into a single Video.
        append(): Append another Video object to the end of the current Video.
        transform(): Apply a transformation function to the individual frames of this object (lazy & in-place).
        iter_frames(): Iterate over the frames in the video from a start point to an end point with a certain step size.
        __iter__(): Makes the Video object an iterable
        __next__(): Get the next frame in the video.
        get_frame(): Get a frame from the video at a specific timestamp or index.
        __getitem__(): Get a frame or a sub-clip or a cropped clip from the video.
        frames_grid(): Create a grid of frames from the video as a single stacked image.
        show(): Display the video.
        show_frame(): Display a specific frame from the video.
        show_frames_grid(): Display a grid of frames from the video as a single stacked image.
        save(): Save the video frames to a file.
        save_frame(): Save a single frame from the video object to the specified path.
        save_frames_grid(): Write a grid of frames from the video to a single image file.
        close(): Release the resources occupied by the object and reset the properties.

    Properties:
        shape: Tuple of array dimensions (n_frames, height, width, n_channels).
        n_frames: The number of frames in the video.
        height: Number of vertical pixels in a video frame.
        width: Number of horizontal pixels in a video frame.
        n_channels: Number of color channels in a video frame.
        duration: Total time that the frames would take to play in a sequence.
        codec: The video codec used to encode the video.

    Example:

        .. code-block:: python

            import sign_language_translator as slt

            # Load a video from a file
            # video = slt.Video("path/to/video.mp4")
            # video = slt.Video("path/to/image.jpg")

            # load from a numpy array
            import numpy as np
            noise = slt.Video(np.random.randint(0, 255, (20,100,160,3), dtype=np.uint8), fps=5)

            # load a dataset file (auto-download)
            video = slt.Video.load_asset("videos/pk-hfad-1_airplane.mp4")
            print(video.duration, video.n_frames)  # 1.72 43

            # trim and concatenate
            video = video + video.trim(start_time=0.5, end_time=1.0)

            # crop
            video = video[:, 100:-100, 50:, :]

            # apply a transformation (flip horizontally)
            video.transform(lambda frame: frame[..., ::-1, :])

            # save & display
            video.save("new_video.mp4", overwrite=True)
            video.save_frames_grid("frames_grid.jpg")
            video.show()
    """

    def __init__(
        self,
        sign: Union[
            str,
            Sequence[Union[NDArray[np.uint8], torch.Tensor]],
            NDArray[np.uint8],
            torch.Tensor,
            Generator[NDArray[np.uint8], None, None],
        ],
        **kwargs,
    ) -> None:
        """Create a new Video object from a video file, a sequence of frames, or a generator of frames.

        Args:
            sign (Union[ str, Sequence[Union[NDArray[np.uint8], torch.Tensor]], NDArray[np.uint8], torch.Tensor, Generator[NDArray[np.uint8], None, None], ]): The video source. Can be a path to a video or image file, a sequence of frames, a single frame, or a generator of frames.
        """
        self._path: Optional[str] = None

        # ToDo: use list of sources instead of linked list of videos.
        # source
        self._source = None
        self.__next: Optional[Video] = None

        # play
        self._source_start_index: int = 0
        # TODO: handle len & end_index for ||step|| > 1
        self._default_step_size: int = 1
        self.transformations: List[Callable] = []

        # properties
        self.fps: float = 30.0
        self.fourcc: int = cv2.VideoWriter_fourcc(*"mp4v")  # type: ignore
        self._height = None
        self._width = None
        self._n_channels = None

        self.__initialize_from_arguments(sign, **kwargs)
        # TODO: from URL
        self._source_end_index = len(self.source) - 1
        self.__update_shape_components()

        # TODO: self.gpu_frame = cv2.cuda_GpuMat()
        self.__idx = 0

    @staticmethod
    def name() -> str:
        return SignFormats.VIDEO.value

    # ========================= #
    #    Iterate / get frame    #
    # ========================= #

    def get_frame(
        self, timestamp: Optional[float] = None, index: Optional[int] = None
    ) -> NDArray[np.uint8]:
        """Get a frame from the video at a specific timestamp or index.

        Args:
            timestamp (Optional[float], optional): The timestamp of the frame to get. If not provided, `index` will be used. If negative, selects the frame from the end of the video. Defaults to None.
            index (Optional[int], optional): The index of the frame to get. If not provided, `timestamp` will be used. if negative, selects the frame from the end of the video (-1 is the last frame). Defaults to None.

        Raises:
            IndexError: If the specified timestamp or index is out of range.

        Returns:
            NDArray[np.uint8]: The 3D RGB frame at the specified timestamp or index from the video.
        """
        # allow negative indexing
        if index and index < 0:
            index = index + len(self)
        if timestamp and timestamp < 0:
            timestamp = timestamp + self.duration

        # validate arguments
        timestamp, _index = _normalize_args_index_and_timestamp(
            timestamp,
            index,
            self.duration if timestamp or self.fps else float("inf"),
            len(self) - 1,
        )

        # get frame
        node, relative_index = self.__get_node(_index)
        if node is not None:
            frame = node.source.get_frame(index=relative_index)
            for transformation in node.transformations:
                frame = transformation(frame)
            return frame

        # error
        raise IndexError(
            "Error reading frame at "
            + (f"{timestamp = }" if index is None else f"{index = }")
        )

    def trim(
        self,
        start_time: Optional[float] = None,
        end_time: Optional[float] = None,
        start_index: Optional[int] = None,
        end_index: Optional[int] = None,
    ) -> Video:
        """Cut the video to a specific time range or index range.
        Allows for negative indexing and negative timestamps which select the frame from the end of the video.
        for example: cut a minute long video with start at 5 sec and end at -15 sec to get a 40 sec video.

        Args:
            start_time (Optional[float], optional): The start time in seconds of the trimmed video. If only `end_time` is provided, trim will start from 0. Defaults to None.
            end_time (Optional[float], optional): The end time in seconds of the trimmed video. If only `start_time` is provided, trim will end at video duration. Defaults to None.
            start_index (Optional[int], optional): The start index of the trimmed video. If only `end_index` is provided, trim will start from 0. Defaults to None.
            end_index (Optional[int], optional): The end index of the trimmed video. If only `start_index` is provided, trim will end at last index (inclusive). Defaults to None.

        Raises:
            ValueError: If the start is not smaller than the end.
            IndexError: If the start index or end index is out of range.

        Returns:
            Video: The trimmed video.
        """

        # allow negative indexing
        if start_index and start_index < 0:
            start_index = start_index + len(self)
        if end_index and end_index < 0:
            end_index = end_index + len(self)
        if start_time and start_time < 0:
            start_time = start_time + self.duration
        if end_time and end_time < 0:
            end_time = end_time + self.duration

        # trim till end if only start is specified
        if start_time is not None and end_index is None and end_time is None:
            end_time = self.duration
        if start_index is not None and end_index is None and end_time is None:
            end_index = len(self) - 1

        # trim from 0 if only end is specified
        if end_time is not None and start_time is None and start_index is None:
            start_time = 0
        if end_index is not None and start_time is None and start_index is None:
            start_index = 0

        # validate arguments
        start_time, start_index = _normalize_args_index_and_timestamp(
            start_time, start_index, self.duration or float("inf"), len(self) - 1
        )
        end_time, end_index = _normalize_args_index_and_timestamp(
            end_time, end_index, self.duration or float("inf"), len(self) - 1
        )
        if start_index > end_index:
            raise ValueError(
                f"invalid cut points. {start_index = } is not smaller than {end_index = }."
            )

        # get nodes to be trimmed
        start_node, relative_start_index = self.__get_node(start_index)
        end_node, relative_end_index = self.__get_node(end_index)
        if not end_node or not start_node:
            raise IndexError(
                f"Error trimming frames at {start_index = }, {end_index = }."
            )

        # trim
        remaining_nodes = end_node.__next
        end_node_source_end_index = end_node._source_end_index
        end_node.__next = None
        end_node._source_end_index = relative_end_index
        new = copy(start_node)
        end_node.__next = remaining_nodes
        end_node._source_end_index = end_node_source_end_index
        new._source_start_index = relative_start_index

        return new

    def iter_frames(
        self, start: int = 0, end: Optional[int] = None, step: Optional[int] = None
    ) -> Generator[NDArray[np.uint8], None, None]:
        """Iterate over the frames in the video from `start` index to `end` index with a certain `step` size.

        Args:
            start (int, optional): The index of the start frame. Defaults to 0.
            end (Optional[int], optional): The index of the end frame. `None` will iterate till the end of the video. Defaults to None.
            step (Optional[int], optional): The step size for iteration. If `None`, uses the default step size of 1. Defaults to None.

        Yields:
            NDArray[np.uint8]: 3D array representing frames from the video with shape: (height, width, color_channels).
        """
        for i in range(start, end or len(self), (step or 1) * self._default_step_size):
            yield self.get_frame(index=i)

    def __get_node(self, index: int) -> Tuple[Optional[Video], int]:
        """
        Find the node in the video linked list that contains the specified index/frame.

        Args:
            index (int): The index to search for within the video linked list.

        Returns:
            Video | None: The node containing the specified index or None if out of range.
        """

        # (0,(5,10),20) + (0,(6,9),10)
        # ....0,5.............6,9.....

        if index < 0:
            return None, -1
        elif index + self._source_start_index <= self._source_end_index:
            return self, index + self._source_start_index
        elif self.__next:
            return self.__next.__get_node(
                index - (self._source_end_index - self._source_start_index + 1)
            )
        else:
            return None, -1

    def __iter__(self):
        self.__idx = 0 if self._default_step_size > 0 else len(self) - 1
        return self

    def __next__(self) -> NDArray[np.uint8]:
        if 0 <= self.__idx < len(self):
            frame = self.get_frame(index=self.__idx)
            self.__idx += self._default_step_size
            return frame

        raise StopIteration

    def __getitem__(
        self, key: Union[int, slice, Sequence[Union[int, slice]]]
    ) -> Union[Video, NDArray[np.uint8]]:
        slices = _validate_and_normalize_slices(key, max_n_dims=4)
        if len(slices) > 4:
            raise ValueError(f"Expected at most 4 slices. Got {len(slices)}: {key}")
        if isinstance(key, int) or (
            isinstance(key, Sequence) and isinstance(key[0], int)
        ):
            return self.get_frame(index=slices[0].start)[slices[1:]]

        time_slice, frame_slices = slices[0], slices[1:]
        new = self.trim(
            start_index=time_slice.start or 0,
            end_index=(
                time_slice.stop - 1 if time_slice.stop is not None else len(self) - 1
            ),
        )

        # TODO: add support for step / reverse video (negative step)
        # every_node._default_step_size *= time_slice.step
        # every_node._source_end_index -= (node._source_end_index - node._source_end_index) % node._default_step_size

        if [fs for fs in frame_slices if fs not in (slice(None), slice(0, None))]:
            new.transform(lambda frame: frame[frame_slices])

        return new

    # ========================= #
    #    Typecast / get data    #
    # ========================= #

    def numpy(self, *args, **kwargs):
        """
        Convert the video frames to a (4D) NumPy array.

        Args:
            *args: Positional arguments to be passed to `np.array()`.
            **kwargs: Keyword arguments to be passed to `np.array()`.

        Returns:
            np.ndarray: A NumPy array representing the video frames.
        """

        return np.array(list(self), *args, **kwargs)

    def __array__(self, *args, **kwargs):
        return self.numpy(*args, **kwargs)

    def torch(self, *args, **kwargs):
        """
        Convert the video frames to a (4D) PyTorch tensor.

        Args:
            *args: Positional arguments to be passed to `torch.tensor()`.
            **kwargs: Keyword arguments to be passed to `torch.tensor()`.

        Returns:
            torch.Tensor: A PyTorch tensor representing the video frames.
        """

        return torch.tensor(self.numpy(), *args, **kwargs)

    # ========================== #
    #    Display / show frame    #
    # ========================== #

    def show(
        self,
        inline_player: Literal["jshtml", "html5"] = "html5",
        codec="avc1",
        **kwargs,
    ) -> None:
        """Display the video. If the function is called from a Jupyter Notebook, the output
        is displayed inline using the specified player. If the video has a single frame,
        it is displayed as an image plot. If the function is called from command line, a
        matplotlib animation window shows the video.

        Note:
            - Clear previous video output in jupyter before displaying the video again to avoid issues.

        Args:
            inline_player (str, optional): The type of matplotlib inline player to use. Defaults to "html5".
            codec (str, optional): The codec to use for the video (should be installed on the system). Possible values are ["avc1", "h264", "mp4v", ...]. Defaults to "avc1".
        """
        # if unmodified
        if (
            in_jupyter_notebook()
            and self._path is not None
            and self.__next is None
            and self._source_start_index == 0
            and self._source_end_index == len(self.source) - 1
            and inline_player == "html5"
            and not self.transformations
        ):
            VideoDisplay.display_ipython_video_in_jupyter(self._path)

        # rewrite
        elif in_jupyter_notebook() and inline_player == "html5" and len(self) > 1:
            if isfile(temp_filename := abspath(join(".", "__temp__.mp4"))):
                remove(temp_filename)
            self.save(temp_filename, overwrite=True, codec=codec)
            VideoDisplay.display_ipython_video_in_jupyter(temp_filename)

        else:
            VideoDisplay.display_frames(
                self, fps=self.fps or 30.0, inline_player=inline_player
            )

    def show_frame(
        self, timestamp: Optional[float] = None, index: Optional[int] = None
    ) -> None:
        """Display a specific frame from the video.

        Args:
            timestamp (Optional[float], optional): The timestamp of the frame to display. If not provided, `index` will be used. Defaults to None.
            index (Optional[int], optional): The index of the frame to display. If not provided, `timestamp` will be used. Defaults to None.
        """
        frame = self.get_frame(timestamp=timestamp, index=index)

        if frame is not None:
            VideoDisplay.display_frames([frame], inline_player="jshtml")

    def frames_grid(
        self,
        rows=2,
        columns=3,
        width: Optional[int] = None,
        height: Optional[int] = None,
    ) -> NDArray[np.uint8]:
        """Create a grid of frames from the video as a single stacked image.
        Equally spaced timestamps are chosen across the video and arranged into a rows x columns grid.

        First frame is placed in the top-right cell. The immediately next frame is placed in the same row and in the column on the right
        and so on until columns in that row run out and then the row below is chosen

        Args:
            rows (int, optional): The number of rows in the grid. Defaults to 2.
            columns (int, optional): The number of columns in the grid. Defaults to 3.
            width (Optional[int], optional): The width of the grid. If only `height` is given, the resized width is calculated by maintaining the aspect ratio of the grid cell. If both are None, the grid is not resized. Defaults to None.
            height (Optional[int], optional): The height of the grid. If only `width` is given, the resized height is calculated by maintaining the aspect ratio of the grid cell. If both are None, the grid is not resized. Defaults to None.

        Returns:
            NDArray[np.uint8]: an RGB 3D numpy array containing the stacked frames. shape: (height, width, color_channels).
        """
        grid = np.concatenate(
            [
                np.concatenate([self.get_frame(index=id) for id in ids_in_row], axis=1)
                for ids_in_row in np.linspace(0, len(self) - 1, rows * columns)
                .reshape(rows, columns)
                .round()
                .astype(int)
            ],
            axis=0,
        )

        if width or height:
            # calculate height or width of grid by maintaining aspect ratio of cell
            aspect_ratio = (self.width * columns) / (self.height * rows)

            if width and not height:
                height = int(width / aspect_ratio)
            elif height and not width:
                width = int(height * aspect_ratio)

            grid = cv2.resize(grid, (width, height))  # type: ignore

        return grid

    def show_frames_grid(
        self,
        rows=2,
        columns=3,
        width: Optional[int] = 800,
        height: Optional[int] = None,
    ):
        """Display a grid of frames from the video as a single stacked image.
        Top left cell contains the first frame. sequence grows towards the right and then down.

        Args:
            rows (int, optional): The number of rows in the grid. Defaults to 2.
            columns (int, optional): The number of columns in the grid. Defaults to 3.
            width (Optional[int], optional): The width of the grid. If only `height` is given, the resized width is calculated by maintaining the aspect ratio of the grid cell. If both are None, the grid is not resized. Defaults to 800.
            height (Optional[int], optional): The height of the grid. If only `width` is given, the resized height is calculated by maintaining the aspect ratio of the grid cell. If both are None, the grid is not resized. Defaults to None.
        """
        grid = self.frames_grid(rows=rows, columns=columns, width=width, height=height)

        VideoDisplay.display_frames([grid], inline_player="jshtml")

    # ================= #
    #    Concatenate    #
    # ================= #

    @staticmethod
    def concatenate(objects: Iterable[Video]) -> Video:
        """
        Concatenate a sequence of Video objects in time dimension into a single Video.

        Args:
            objects (Iterable[Video]): A sequence of Video objects to concatenate.

        Returns:
            Video: A new Video object that is a linked list of all the input videos.

        Raises:
            ValueError: If the input sequence of videos is empty.
        """

        video_chain = None
        for video in objects:
            if video_chain is None:
                video_chain = deepcopy(video)
            else:
                video_chain.append(video)

        if video_chain is None:
            raise ValueError("Cannot concatenate an empty sequence of videos.")

        return video_chain

    @classmethod
    def stack(
        cls,
        videos: Sequence[Video],
        dim=1,
        fps: Union[
            float, Literal["first", "max", "average", "weighted", "min"]
        ] = "max",
    ):
        """
        Stack a list of Video objects along a specified dimension (0=time, 1=height, 2=width, 3=channels).
        All videos will be aligned at the start and padded with black frames at the end if necessary.

        Args:
            videos (Sequence[Video]): A list of Video objects to stack.
            dim (int, optional): The dimension along which to stack the videos (0=time, 1=height, 2=width, 3=channels). Negative values represent dimensions relative to the end (e.g., -1 is the channel dimension). Defaults to 1.
            fps (Union[float, Literal["first", "max", "average", "weighted"]], optional): The frames per second of the output video. If "first", it will use the fps of the first video. If "max", it will use the maximum fps of all videos. If "average", it will use the average fps of all videos. If "weighted", it will use the weighted average fps of all videos (weights=duration). If "min", it will use the minimum fps of all videos.  Not used for time dimension. Defaults to "max".

        Returns:
            Video: A new Video object resulting from stacking the input videos.
        """

        # --- validate arguments --- #

        if dim < -4 or dim > 3:
            raise ValueError(f"Dimension should be in range -4<=dim<=3. Got {dim = }")
        if dim < 0:
            dim = dim + 4
        if dim == 0:
            return cls.concatenate(videos)

        if fps == "first":
            fps = videos[0].fps
        elif fps == "max":
            fps = max(v.fps for v in videos)
        elif fps == "min":
            fps = min(v.fps for v in videos)
        elif fps == "average":
            fps = sum(v.fps for v in videos) / len(videos)
        elif fps == "weighted":
            fps = sum(len(v) for v in videos) / sum(v.duration for v in videos)
        if not isinstance(fps, (float, int)):
            raise ValueError(f"Invalid value for {fps = }")

        # --- stack videos --- #

        def stacked_frames(videos: Sequence[Video], fps: float):
            duration = max(v.duration for v in videos)

            for t in np.arange(0.5 / fps, duration, 1 / fps):
                frames = []
                for video in videos:
                    if t <= video.duration:
                        frames.append(video.get_frame(timestamp=t))
                    else:
                        frames.append(np.zeros(video.shape[1:], dtype=np.uint8))

                # todo: resize frames to the same height/width while maintaining aspect ratio
                yield np.concatenate(frames, axis=dim - 1)

        return cls(
            stacked_frames(videos, fps),
            fps=fps,
            total_frames=int(max(v.duration for v in videos) * fps + 0.5),
        )

    def append(self, other: Video):
        """
        Append another Video object to the end of the current Video.

        Args:
            other (Video): The Video object to append.

        Returns:
            None
        """

        # prevent cycle in the video linked list
        other = copy(other)

        if self.__next is None:
            self.__next = other
        else:
            self.__next.append(other)

    def __add__(self, other: Video):
        new = copy(self)
        new.append(copy(other))
        return new

    def __copy__(self):
        new = self.__class__.__new__(self.__class__)
        new.__dict__.update({k: copy(v) for k, v in self.__dict__.items()})
        new._source_end_index = self._source_end_index

        return new

    # =============== #
    #    Transform    #
    # =============== #

    def transform(
        self, transformation: Callable[[NDArray[np.uint8]], NDArray[np.uint8]]
    ):
        """
        Apply a transformation function to the individual frames of this object (in place).

        Args:
            transformation (Callable[[NDArray[np.uint8]], NDArray[np.uint8]]): The transformation function to be applied that must map 1 numpy array to another numpy array.

        Raises:
            ValueError: If the transformation is not callable.

        Returns:
            None
        """

        if not isinstance(transformation, (Callable,)):
            raise ValueError(
                f"Cannot apply this transformation because it is not callable. {transformation = }"
            )

        self.transformations.append(transformation)
        if self.__next:
            self.__next.transform(transformation)

        self.__update_shape_components()

    # ================ #
    #    Dimensions    #
    # ================ #

    def __len__(self) -> int:
        # TODO: handle steps e.g. video[::3, ...]
        # todo: fix len when transformation is applied
        return (self._source_end_index - self._source_start_index + 1) + len(
            self.__next or ()
        )

    @property
    def shape(self) -> Tuple[int, int, int, int]:
        """Tuple of array dimensions (n_frames, height, width, n_channels)."""
        return len(self), self.height, self.width, self.n_channels

    @property
    def height(self) -> int:
        """number of vertical pixels in a video frame (dimension=1)"""
        if self._height is None:
            raise ValueError("self._height is not defined.")
        return self._height

    @property
    def width(self) -> int:
        """number of horizontal pixels in a video frame (dimension=2)"""
        if self._width is None:
            raise ValueError("self._width is not defined.")
        return self._width

    @property
    def n_channels(self) -> int:
        """number of color channels in a video frame (e.g. RGB) (dimension=3)"""
        if self._n_channels is None:
            raise ValueError("self._n_channels is not defined.")
        return self._n_channels

    @property
    def duration(self) -> float:
        """total time that the frames would take to play in a sequence. depends on fps."""
        if self.fps is None:
            # TODO: remove this
            raise RuntimeError("FPS is not set. Can not get duration without fps.")

        return len(self) / self.fps if self.fps else float("inf")

    @property
    def n_frames(self) -> int:
        """The number of frames in the video."""
        return len(self)

    # ================== #
    #    Save to disk    #
    # ================== #

    @staticmethod
    def save_(  # TODO: overload save to be static and instance method simultaneously
        frames_iterable: Iterable[NDArray[np.uint8]],
        path: str,
        overwrite=False,
        height: Optional[int] = None,
        width: Optional[int] = None,
        fps: float = 30.0,
        codec: str = "mp4v",
        progress_bar=True,
        leave=False,
        total_frames: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Class method to save a sequence of frames into a video file.
        A frame is a 3D numpy uint8 array of shape (height, width, channels)

        Args:
            path (str): The path to the output file.
            overwrite (bool, optional): Whether to overwrite the output file if it already exists. Defaults to False.
            height (Optional[int], optional): The number of vertical pixels in the output video file. If None, the shape (index=0) of the first frame is used. Defaults to None.
            width (Optional[int], optional): The number of horizontal pixels in the output video file. If None, the shape (index=1) of the first frame is used. Defaults to None.
            fps (float, optional): The frames per second of the output video. Defaults to 30.0.
            codec (str, optional): The codec used for the output video e.g ["h264", "mp4v", "xvid", "avc1", "hvc1"]. Make sure the codec is already installed in your system (some do not ship with OpenCV because of license mismatch). Defaults to "mp4v".
            progress_bar (bool, optional): Whether to display a progress bar while saving the video. Defaults to True.
            leave (bool, optional): Whether to leave the progress bar after the operation is complete. Defaults to False.
            total_frames (Optional[int], optional): total number of frames in the frames iterable. Used by the progress bar. Defaults to None.

        Raises:
            FileExistsError: If a file already exists at the output path and `overwrite` is False.
        """
        path = validate_path_exists(path, overwrite=overwrite)

        _frames_iterable = iter(
            frames_iterable
            if not progress_bar
            else tqdm(
                frames_iterable,
                total=getattr(frames_iterable, "__len__", None) or total_frames,
                leave=leave,
            )
        )
        first_frame = next(_frames_iterable)
        height = height or first_frame.shape[0]
        width = width or first_frame.shape[1]

        # TODO: Compile opencv with GPU support and write to disk form CUDA.
        writer = cv2.VideoWriter(
            path, cv2.VideoWriter_fourcc(*codec), fps, (width, height)  # type: ignore
        )

        try:
            writer.write(cv2.cvtColor(first_frame, cv2.COLOR_BGR2RGB))
            for frame in _frames_iterable:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                writer.write(frame)
        finally:
            writer.release()

    def save(
        self,
        path: str,
        overwrite=False,
        fps: Optional[float] = None,
        codec: Optional[str] = None,
        progress_bar=True,
        leave=False,
        **kwargs,
    ) -> None:
        """Save the video frames to a file.

        Args:
            path (str): The path to the output file.
            fps (float | None, optional): The frames per second of the output video. If None, it will use the fps of the video source or 30 if that is not set either. Defaults to None.
            codec (str | None, optional): The codec used for the output video e.g ["h264", "mp4v", "xvid", "avc1", "hvc1"]. Make sure the codec is already installed in your system (some do not ship with OpenCV because of license mismatch). If None, it will use the source video codec or "mp4v" if that is not set. Defaults to None.
            overwrite (bool, optional): Whether to overwrite the output file if it already exists. Defaults to False.
            progress_bar (bool, optional): Whether to display a progress bar while saving the video. Defaults to True.
            leave (bool, optional): Whether to leave the progress bar after the operation is complete. Defaults to False.
            **kwargs: Additional keyword arguments. (not used yet.)
        """

        self.save_(
            self.iter_frames(),
            path,
            overwrite=overwrite,
            width=self.width,
            height=self.height,
            fps=fps or self.fps or 30,
            codec=codec or self.codec or "mp4v",
            progress_bar=progress_bar,
            leave=leave,
            total_frames=len(self),
            **kwargs,
        )

    def save_frame(
        self,
        path: str,
        timestamp: Optional[float] = None,
        index: Optional[int] = None,
        overwrite=False,
    ) -> None:
        """
        Saves a single frame from the video object to the specified path.

        Args:
            path (str): The path where the frame will be saved.
            timestamp (float | None, optional): The timestamp of the frame to be saved. If None, `index` will be used. Defaults to None.
            index (int | None, optional): The index of the frame to be saved. If None, `timestamp` will be used. Defaults to None.
            overwrite (bool, optional): Whether to overwrite the image file if it already exists. Defaults to False.

        Raises:
            ValueError: If both or neither `timestamp` and `index` are provided or If the specified timestamp or index is out of range.
            FileExistsError: If a file already exists at the output path and `overwrite` is False.
        """
        path = validate_path_exists(path, overwrite=overwrite)

        frame = self.get_frame(timestamp=timestamp, index=index)
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        cv2.imwrite(path, frame)

    def save_frames_grid(
        self,
        path: str,
        rows: int = 2,
        columns: int = 3,
        width: Optional[int] = 1024,
        height: Optional[int] = None,
        overwrite=False,
    ) -> None:
        """Write a grid of frames from the video to a single image file.
        The grid is created by arranging frames from the video taken at
        equally spaced timestamps in a rows x columns grid.

        Args:
            path (str): The path to the output image file.
            rows (int, optional): The number of rows in the grid. Defaults to 2.
            columns (int, optional): The number of columns in the grid. Defaults to 3.
            width (Optional[int], optional): The width of the grid. If only `height` is given, the resized width is calculated by maintaining the aspect ratio of the grid cell. If both are None, the grid is not resized. Defaults to 1024.
            height (Optional[int], optional): The height of the grid. If only `width` is given, the resized height is calculated by maintaining the aspect ratio of the grid cell. If both are None, the grid is not resized. Defaults to None.
            overwrite (bool, optional): Whether to overwrite the image file if it already exists. Defaults to False.

        Raises:
            FileExistsError: If a file already exists at the output path and `overwrite` is False.
        """
        path = validate_path_exists(path, overwrite=overwrite)

        grid = self.frames_grid(rows=rows, columns=columns, width=width, height=height)
        grid = cv2.cvtColor(grid, cv2.COLOR_RGB2BGR)
        cv2.imwrite(path, grid)

    # ============================= #
    #    Initialize Video object    #
    # ============================= #

    @classmethod
    def load(cls, path: str, **kwargs) -> Video:
        """
        Load a video from the specified path.

        Args:
            path (str): The path to the video file.
            **kwargs: Additional keyword arguments to be passed to the Video constructor.

        Returns:
            Video: The loaded video object.
        """
        return cls(path, **kwargs)

    @classmethod
    def load_asset(
        cls,
        label: str,
        archive_name: Optional[str] = None,
        overwrite=False,
        progress_bar=True,
        leave=True,
        **kwargs,
    ) -> Video:
        """Class method to load a video asset identified by the given label from
        the built-in datasets and return it as a new Video object.

        This method downloads dictionary videos from direct URLs if the corresponding
        archive is not already downloaded. Otherwise the video asset will be extracted
        from an archive which will be auto-downloaded once.

        Notes:
        - To view valid asset IDs, run `slt.Assets.get_ids(r"^videos/.*mp4$")` for dictionary videos
          or `slt.Assets.get_ids(r"^datasets/.*videos.*zip$")` for archived videos.
        - See `slt.Assets.ROOT_DIR` for download directory

        Args:
            label (str): The filename of the video asset to load. 'videos/' is prepended to the label if it does not start with it and '.mp4' is appended to the label if it does not end with it. An example is `videos/pk-hfad-1_airplane.mp4` for a dictionary video. General syntax is `videos/country-organization-number_text[_person_camera].mp4`.
            archive_name (str | None, optional): The name of the archive which contains the video asset. If None, the archive name is inferred from the label. Not necessary for dictionary videos. An example is `datasets/pk-hfad-1.videos-mp4.zip`. General syntax is `datasets/country-organization-number[_person_camera].videos-mp4.zip`. Defaults to None.
            overwrite (bool, optional): Whether to overwrite the video asset if it is already downloaded or extracted. Defaults to False.
            progress_bar (bool, optional): Whether to display a progress bar while downloading or extracting the asset. Defaults to True.
            leave (bool, optional): Whether to leave the progress bar after the operation is complete. Defaults to True.
            **kwargs: Additional keyword arguments to be passed to the Video constructor.

        Raises:
            FileNotFoundError: If no video assets are found for the given label.

        Warns:
            UserWarning: If multiple video assets match the given label and the only first asset is used.

        Returns:
            Video: An instance of the Video class representing the video that matched the label.

        Examples:

            .. code-block:: python

                import sign_language_translator as slt

                # Load a dictionary video asset
                video = slt.Video.load_asset("pk-hfad-1_airplane")

                # Load a replication video from the built-in datasets
                video = slt.Video.load_asset("videos/pk-hfad-1_airplane_dm0001_front.mp4", archive_name="datasets/pk-hfad-1_dm0001_front.videos-mp4.zip")
        """

        if "/" not in label:
            label = f"videos/{label}"

        if "." not in label:
            label = f"{label}.mp4"

        paths = []
        if Assets.is_dictionary_video(label):
            # if no corresponding archive is already downloaded
            if not any(
                isfile(archive_path)
                for archive_path in Assets.get_path(
                    archive_name or Assets.infer_archive_name(label)
                )
            ):
                # download video from direct URL
                paths = Assets.download(
                    label, progress_bar=progress_bar, leave=leave, overwrite=overwrite
                )

        # extract video from archive
        if not paths:
            paths = Assets.extract(
                label,
                archive_name_or_regex=archive_name,
                download_archive=True,
                overwrite=overwrite,
                progress_bar=progress_bar,
                leave=leave,
            )

        if len(paths) == 0:
            raise FileNotFoundError(f"No video assets found for '{label}'")
        if len(paths) > 1:
            warn(f"Multiple video assets matched '{label}'. Using the first of:{paths}")

        return cls.load(paths[0], **kwargs)

    def __initialize_from_arguments(self, sign, **kwargs):
        if isinstance(sign, str):
            if not isfile(sign):
                raise ValueError(f"Invalid path argument: '{sign}'")
            self._from_path(sign, **kwargs)

        elif isinstance(sign, (np.ndarray, torch.Tensor, Sequence)):
            self._from_frames(sign, **kwargs)  # type: ignore
        elif isinstance(sign, Iterable):
            self._from_iterable(
                sign,
                kwargs["total_frames"],
                **{k: v for k, v in kwargs.items() if k not in ("total_frames",)},
            )

        else:
            raise ValueError(
                f"Invalid argument: {sign}. provide either a path to a video or a sequence of frames."
            )

    def _from_path(self, path: str, **kwargs) -> None:
        self._path = abspath(path)
        file_type = str(guess_type(path)[0])

        if file_type.startswith("image"):
            self._from_frames(
                [cv2.cvtColor(cv2.imread(path), cv2.COLOR_BGR2RGB)],  # type: ignore
                fps=kwargs.get("fps", 1),
                **kwargs,
            )

        elif file_type.startswith("video"):
            self.source = VideoCaptureFrames(self._path, **kwargs)
            self.__extract_properties_from_source(self.source)

        else:
            raise ValueError(f"unknown file type: {file_type}")

    def __extract_properties_from_source(self, source: VideoFrames) -> None:
        self.fps = getattr(source, "fps", 30.0)
        self.fourcc = getattr(source, "fourcc", cv2.VideoWriter_fourcc(*"mp4v"))  # type: ignore

    @property
    def codec(self) -> str:
        """The video codec used to encode the video. (e.g. "mp4v", "h264", "xvid", "avc1", "hvc1")"""
        return struct.pack("<I", self.fourcc).decode("utf-8")

    def __update_shape_components(self) -> None:
        shape = self.get_frame(0).shape
        self._height = shape[0]
        self._width = shape[1]
        self._n_channels = shape[2]

    def _from_data(self, sign_data: Sequence[NDArray[np.uint8]], **kwargs) -> None:
        if isinstance(sign_data, Sequence):
            self._from_frames(sign_data, **kwargs)
        else:
            self._from_iterable(sign_data, kwargs["total_frames"], **kwargs)

    def _from_frames(
        self, frames: Sequence[NDArray[np.uint8]], fps: float = 30.0, **kwargs
    ):
        # validation
        if len(frames) <= 0:
            raise ValueError("Can not initialize from None or empty frames argument.")
        if frames[0].ndim != 3:
            raise ValueError("Each frame must be 3D (Height, Width, Channels).")
        if isinstance(frames, torch.Tensor):
            frames = frames.numpy().astype(np.uint8)
        elif isinstance(frames[0], torch.Tensor):
            frames = torch.Tensor(frames).numpy().astype(np.uint8)

        # source
        self.source = SequenceFrames(frames, fps=fps)
        self.__extract_properties_from_source(self.source)

    def _from_iterable(self, frames_iterable: Iterable, total_frames: int, **kwargs):
        """load the frames from a read-once iterable such as a generator."""
        self.source = IterableFrames(frames_iterable, total_frames, **kwargs)
        self.__extract_properties_from_source(self.source)

    @property
    def source(self) -> VideoFrames:
        """A VideoFrames object which wraps around a frame sequence."""
        if self._source is None:
            raise ValueError("Video source is not set.")
        return self._source

    @source.setter
    def source(self, source: VideoFrames):
        if not isinstance(source, VideoFrames):
            raise ValueError(
                f"Invalid source type: {type(source)}. Should be slt.vision.video.VideoFrames"
            )
        self._source = source

    # ============================= #
    #    Cleaning / with _ as _:    #
    # ============================= #

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        """Release the resources occupied by the object and reset the properties."""
        # source
        if self._source is not None:
            self.source.close()
        if self.__next is not None:
            self.__next.close()
            self.__next = None

        # play
        self._source_start_index = 0
        self._source_end_index = 0
        self.transformations = []

        # properties
        self._path = None

        self.is_closed = True



__all__ = [
    "BaseConnections",
    "MediapipeConnections",
    "get_connections",
]

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Sequence, Tuple

from sign_language_translator.config.colors import Colors


class Connection:
    """
    Represents a connection between landmarks.

    Attributes:
        label (str): The label of the connection.
        indexes (Sequence[int]): The indexes of the landmarks connected by this connection.
        color (Tuple[int, int, int]): The color of the connection.
    """

    def __init__(
        self, label: str, indexes: Sequence[int], color: Tuple[int, int, int]
    ) -> None:
        self.label = label
        self.indexes = indexes
        self.color = color

    def __repr__(self) -> str:
        return f"Connection(label={self.label}, indexes={self.indexes}, color={self.color})"


class BaseConnections(ABC):
    """A class containing information about the connections between landmarks generated from various models."""

    def __init__(self) -> None:
        self._line_indexes = []
        self._line_colors = []
        self._line_labels = []
        self._matplot3d_config = {}

    @staticmethod
    @abstractmethod
    def name() -> str:
        """The name of the connection format"""

    @property
    @abstractmethod
    def connections(self) -> List[Connection]:
        """indexes of landmarks that are connected"""

    @property
    @abstractmethod
    def n_landmarks(self) -> int:
        """Total number of landmarks"""

    @property
    @abstractmethod
    def n_features(self) -> int:
        """Total number of features per landmark"""

    @property
    def line_indexes(self) -> List[Sequence[int]]:
        """list of sequence of indexes that are connected with single line"""
        return self._line_indexes

    @property
    def line_colors(self) -> List[Tuple[int, int, int]]:
        """list of colors for each connection"""
        return self._line_colors

    @property
    def line_labels(self) -> List[str]:
        """list of labels for each connection"""
        return self._line_labels

    @property
    def matplot3d_config(self) -> Dict[str, Any]:
        """Configuration arguments for 3D matplotlib plot"""
        return self._matplot3d_config

    def _parse_connections(self) -> None:
        """extract line indexes, colors and labels from connections into separate lists"""
        self._line_indexes = [connection.indexes for connection in self.connections]
        self._line_colors = [connection.color for connection in self.connections]
        self._line_labels = [connection.label for connection in self.connections]

    # left hand range
    # right hand range
    # left hand connections
    # right hand connections
    # torso (for rectification)


class MediapipeConnections(BaseConnections):
    """Represents the connections for the Mediapipe landmark model."""

    def __init__(self) -> None:
        super().__init__()

        self.scatter_color = Colors.BLACK
        self.default_color = Colors.BLACK
        self.default_color_2 = Colors.DARK_GREY
        self._n_landmarks = 75  # 33 + 21 * 2
        self._n_features = 5

        # Initialize all the connections
        self.face = Connection(
            "face", [8, 6, 5, 4, 0, 10, 9, 0, 1, 2, 3, 7], Colors.BLACK
        )
        self.torso = Connection("torso", [12, 11, 23, 24, 12], Colors.BLACK)
        self.right_arm = Connection(
            "right_arm", [12, 14, 16, 18, 20, 16, 22], Colors.MAGENTA
        )
        self.left_arm = Connection(
            "left_arm", [11, 13, 15, 17, 19, 15, 21], Colors.CYAN
        )
        self.right_leg = Connection("right_leg", [24, 26, 28, 30, 32, 28], Colors.BLACK)
        self.left_leg = Connection("left_leg", [23, 25, 27, 29, 31, 27], Colors.BLACK)
        self.left_palm = Connection(
            "left_palm", [38, 42, 46, 50, 33, 38, 34], Colors.BLACK
        )
        self.left_thumb = Connection("left_thumb", [33, 34, 35, 36, 37], Colors.GREEN)
        self.left_index = Connection("left_index", [38, 39, 40, 41], Colors.DARK_GREY)
        self.left_middle = Connection("left_middle", [42, 43, 44, 45], Colors.BLACK)
        self.left_ring = Connection("left_ring", [46, 47, 48, 49], Colors.BLACK)
        self.left_pinky = Connection("left_pinky", [50, 51, 52, 53], Colors.BLUE)
        self.right_palm = Connection(
            "right_palm", [59, 63, 67, 71, 54, 59, 55], Colors.BLACK
        )
        self.right_thumb = Connection(
            "right_thumb", [54, 55, 56, 57, 58], Colors.ORANGE
        )
        self.right_index = Connection("right_index", [59, 60, 61, 62], Colors.GREY)
        self.right_middle = Connection("right_middle", [63, 64, 65, 66], Colors.BLACK)
        self.right_ring = Connection("right_ring", [67, 68, 69, 70], Colors.BLACK)
        self.right_pinky = Connection("right_pinky", [71, 72, 73, 74], Colors.HOT_PINK)

        # Add all the connections to the list property
        self._connections = [
            self.face,
            self.torso,
            self.right_arm,
            self.left_arm,
            self.right_leg,
            self.left_leg,
            self.left_palm,
            self.left_thumb,
            self.left_index,
            self.left_middle,
            self.left_ring,
            self.left_pinky,
            self.right_palm,
            self.right_thumb,
            self.right_index,
            self.right_middle,
            self.right_ring,
            self.right_pinky,
        ]
        self._parse_connections()

        # Set the configuration for Matplotlib 3D visualization
        self._matplot3d_config = {
            "line_indexes": self.line_indexes,
            "line_colors": [tuple(v / 255 for v in c) for c in self.line_colors],
            "line_labels": self.line_labels,
            "invert_y": True,
            "invert_z": True,
            "vertical_axis": "y",
            "scatter_color": tuple(v / 255 for v in self.scatter_color),
        }

    @staticmethod
    def name() -> str:
        return "mediapipe"

    @property
    def connections(self) -> List[Connection]:
        return self._connections

    @property
    def n_landmarks(self) -> int:
        return self._n_landmarks

    @property
    def n_features(self) -> int:
        return self._n_features


def get_connections(connections: str) -> BaseConnections:
    """Create a connections object based on the given string

    Args:
        connections (str): The name of the connections format to use.

    Returns:
        BaseConnections: The connections object.

    Raises:
        ValueError: If the connections format is not recognized.
    """

    if connections in ("mediapipe-world", "mediapipe-image"):
        return MediapipeConnections()

    raise ValueError(f"Unknown connections format: {connections}")



from __future__ import annotations

__all__ = [
    "Sign",
]

from abc import ABC, abstractmethod
from typing import Callable, Iterable, Optional, Union

import torch
from numpy.typing import NDArray


class Sign(ABC):
    def __init__(self, sign, *args, **kwargs) -> None:
        """initialize the sign"""

    @staticmethod
    @abstractmethod
    def name() -> str:
        """the string code of the sign format"""

    @abstractmethod
    def show(self, **kwargs) -> None:
        """display the sign"""

    @abstractmethod
    def _from_path(self, path: str):
        """load the sign from a path"""

    @abstractmethod
    def _from_data(self, sign_data):
        """load the sign from an object"""

    @classmethod
    @abstractmethod
    def load(cls, path: str, **kwargs) -> Sign:
        """read the sign from a path"""

    @classmethod
    @abstractmethod
    def load_asset(
        cls, label: str, archive_name: Optional[str] = None, **kwargs
    ) -> Sign:
        """load a sign asset from a dataset"""

    @staticmethod
    @abstractmethod
    def concatenate(objects: Iterable[Sign]) -> Sign:
        """concatenate multiple signs in time dimension"""

    @abstractmethod
    def numpy(self, *args, **kwargs) -> NDArray:
        """return the sign as a numpy array"""

    @abstractmethod
    def torch(
        self,
        dtype: Optional[torch.dtype] = None,
        device: Union[torch.device, str, None] = None,
    ) -> torch.Tensor:
        """return the sign as a torch tensor"""

    @abstractmethod
    def transform(self, transformation: Callable):
        """apply some transformation to the sign to change its appearance"""

    @abstractmethod
    def save(self, path: str, **kwargs) -> None:
        """save the sign to a path"""



__all__ = [
    "Colors",
]

from dataclasses import dataclass
from typing import List, Sequence, Tuple, Union

import numpy as np

from sign_language_translator.utils.arrays import linear_interpolation


@dataclass(frozen=True)
class Colors:
    """A class to represent a set of predefined colors and provide utility functions related to colors."""

    BLACK: Tuple[int, int, int] = (0, 0, 0)
    NAVY_BLUE: Tuple[int, int, int] = (0, 0, 255)
    GREEN: Tuple[int, int, int] = (0, 255, 0)
    RED: Tuple[int, int, int] = (255, 0, 0)
    CYAN: Tuple[int, int, int] = (0, 255, 255)
    MAGENTA: Tuple[int, int, int] = (255, 0, 255)
    YELLOW: Tuple[int, int, int] = (255, 255, 0)
    ORANGE: Tuple[int, int, int] = (255, 127, 0)
    PURPLE: Tuple[int, int, int] = (127, 0, 255)
    PINK: Tuple[int, int, int] = (255, 127, 255)
    HOT_PINK: Tuple[int, int, int] = (255, 0, 127)
    BLUE: Tuple[int, int, int] = (0, 127, 255)
    LEAF_GREEN: Tuple[int, int, int] = (127, 255, 0)
    DARK_GREY: Tuple[int, int, int] = (63, 63, 63)
    GREY: Tuple[int, int, int] = (127, 127, 127)
    LIGHT_GREY: Tuple[int, int, int] = (191, 191, 191)

    @staticmethod
    def gradient(
        n: int,
        colors: Sequence[
            Tuple[Union[float, int], Union[float, int], Union[float, int]]
        ] = (HOT_PINK, BLUE, HOT_PINK),
        endpoint=False,
        dtype=int,
    ) -> List[Tuple[float, float, float]]:
        """
        Generate a gradient of colors by linearly interpolating between a sequence of colors.

        Args:
            n (int): The number of colors to generate in the gradient.
            colors (Sequence[Tuple[Union[float, int], Union[float, int], Union[float, int]]], optional): A sequence of RGB tuples to interpolate between. Defaults to (HOT_PINK, BLUE, HOT_PINK).
            endpoint (bool, optional): If True, the last value in the colors sequence is included. Defaults to False.
            dtype (Union[type, str], optional): The desired data type of the output colors. Defaults to int.

        Returns:
            List[Tuple[float, float, float]]: A list of RGB tuples representing the gradient.

        Example:
            >>> Colors.gradient(5, colors = [(128, 0, 0), (0, 0, 128)], endpoint=True)
            [[128, 0, 0], [96, 0, 32], [64, 0, 64], [32, 0, 96], [0, 0, 128]]
        """
        return (
            np.array(
                linear_interpolation(
                    colors, np.linspace(0, len(colors) - 1, n, endpoint=endpoint)
                )
            )
            .astype(dtype)
            .tolist()
        )



import math
import re
from functools import partial
from itertools import zip_longest
from typing import List, Literal, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FuncAnimation
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.ticker import MultipleLocator
from mpl_toolkits.mplot3d.art3d import Line3D, Path3DCollection
from numpy.typing import NDArray

from sign_language_translator.config.colors import Colors


class MatPlot3D:

    @staticmethod
    def new_figure(
        x_limits: Tuple[float, float],
        y_limits: Tuple[float, float],
        z_limits: Tuple[float, float],
        vertical_axis: Literal["x", "y", "z"] = "z",
        figure_scale: Optional[float] = 5,
        style: Literal["dark_background", "default"] = "default",
        layout: Literal["constrained", "compressed", "tight", "none"] = "compressed",
        subplots: Tuple[int, int] = (1, 1),
    ) -> Tuple[Figure, List[Axes]]:
        """Creates a new 3D figure with the specified subplots and settings."""
        # figure size
        if figure_scale and figure_scale > 0:
            sizes: List = np.stack([x_limits, y_limits, z_limits]).ptp(1).tolist()
            ratio = sizes.pop("xyz".index(vertical_axis.lower()))
            ratio = ratio / np.linalg.norm(sizes)
            ratio = 0.9 * ratio  # make it wider
            figsize = (
                figure_scale * max(1, 1 / ratio) * subplots[1],
                figure_scale * max(1, ratio) * subplots[0],
            )
        else:
            figsize = None

        plt.style.use(style)  # todo: set only for this figure
        fig = plt.figure(figsize=figsize, layout=layout)
        axes = [
            fig.add_subplot(*subplots, i + 1, projection="3d")
            for i in range(np.prod(subplots))
        ]

        return fig, axes

    @staticmethod
    def initialize_Axes3D(
        ax: Axes,
        x_limits: Tuple[float, float],
        y_limits: Tuple[float, float],
        z_limits: Tuple[float, float],
        ticks_scale: float = 1.0,
        azimuth: float = 20,
        elevation: float = 15,
        roll: float = 0,
        vertical_axis: str = "y",
        invert_x: bool = False,
        invert_y: bool = False,
        invert_z: bool = False,
        show_grid: bool = True,
        show_axis: bool = True,
    ) -> None:
        """Initializes a 3D Axes object with specified limits, ticks, and settings.

        Args:
            ax (Axes): The 3D Axes object to be initialized.
            x_limits (Tuple[float, float]): The range of the x-axis from minimum to maximum value.
            y_limits (Tuple[float, float]): The range of the y-axis from minimum to maximum value.
            z_limits (Tuple[float, float]): The range of the z-axis from minimum to maximum value.
        """
        ax.set(xlim3d=x_limits, xlabel="X")
        ax.set(ylim3d=y_limits, ylabel="Y")
        ax.set(zlim3d=z_limits, zlabel="Z")

        if not ticks_scale:
            ticks_scale = _nearest_scale(
                np.array([x_limits, y_limits, z_limits]).ptp(1).min()
            )

        ax.xaxis.set_major_locator(MultipleLocator(ticks_scale))
        ax.yaxis.set_major_locator(MultipleLocator(ticks_scale))
        ax.zaxis.set_major_locator(MultipleLocator(ticks_scale))  # type: ignore

        if invert_x:
            ax.invert_xaxis()
        if invert_y:
            ax.invert_yaxis()
        if invert_z:
            ax.invert_zaxis()  # type: ignore

        ax.set_box_aspect(
            np.array([x_limits, y_limits, z_limits]).ptp(1)[
                # change aspect ratio order according to vertical axis
                {"x": [1, 2, 0], "y": [2, 0, 1], "z": [0, 1, 2]}[vertical_axis.lower()]
            ]
        )

        ax.view_init(  # type: ignore
            azim=azimuth, elev=elevation, roll=roll, vertical_axis=vertical_axis
        )
        ax.grid(show_grid)
        ax.axis("on" if show_axis else "off")

    @staticmethod
    def placeholder_scatter_and_lines(
        ax: Axes,
        n_lines: int,
        line_colors: Sequence[Union[Tuple[float, float, float], None]] = (),
        line_labels: Sequence[Union[str, None]] = (),
        scatter_color: Tuple[float, float, float] = (0, 0, 0),
        scatter_size: float = 2,
    ) -> Tuple[Path3DCollection, List[Line3D]]:
        """
        Update a 3D plot with empty Path3DCollection (scatter) and Line3D objects.

        Args:
            ax (Axes): The 3D axes object to plot on.
            n_lines (int): The number of placeholder lines to create.
            line_colors (Sequence[Union[Tuple[float, float, float], None]], optional): The colors of the lines. If not provided, a gradient of colors will be used.
            line_labels (Sequence[Union[str, None]], optional): The labels for the lines.
            scatter_color (Tuple[float, float, float], optional): The RGB color of the scatter points normalized to [0.0, 1.0] range. Defaults to black.
            scatter_size (float, optional): The size of the scatter points. Defaults to 2.

        Returns:
            Tuple[Path3DCollection, List[Line3D]]: A tuple containing the scatter plot and the list of lines.
        """
        n_lines = max(round(n_lines), 0)
        if not line_colors:
            line_colors = np.array(Colors.gradient(n_lines)) / 255  # type: ignore

        scatter = ax.scatter3D(xs=[], ys=[], zs=[], color=scatter_color, s=scatter_size)  # type: ignore
        lines = [
            ax.plot3D(xs=[], ys=[], zs=[], color=color, label=label)[0]  # type: ignore
            for i, color, label in zip_longest(range(n_lines), line_colors, line_labels)
            if i < n_lines
        ]
        return scatter, lines

    @staticmethod
    def set_frame_data(
        points: Union[Sequence[Tuple[float, float, float]], NDArray],
        scatter: Path3DCollection,
        lines: Sequence[Line3D],
        line_indexes: Sequence[Sequence[int]] = (),
        ax: Optional[Axes] = None,
        azimuth_delta: float = 0,
        elevation_delta: float = 0,
        roll_delta: float = 0,
    ) -> List[Union[Path3DCollection, Line3D]]:
        """Sets the frame data for visualization.

        Args:
            points (Union[Sequence[Tuple[float, float, float]], NDArray]): A collection of tuples or a 2D NDArray representing the (x, y, z) points.
            scatter (Path3DCollection): Object representing the scatter plot.
            lines (Sequence[Line3D]): A sequence of Line3D objects representing the lines to be plotted.
            line_indexes (Sequence[Sequence[int]], optional): indexes of points to connect with lines. Defaults to ().
            ax (Optional[Axes], optional): An optional Axes object to update the view. Defaults to None.

        Returns:
            List[Union[Path3DCollection, Line3D]]: A list containing the updated scatter plot and lines objects.
        """
        if len(line_indexes) == 0 and lines:
            line_indexes = _indexes_to_connect(len(points))

        # update view
        if ax is not None:
            if ax.get_title():
                ax.set_title(
                    re.sub(
                        r"\[t_-?\d+\]",
                        lambda timestamp: f"[t_{int(timestamp.group()[3:-1])+1}]",
                        ax.get_title(),
                    )
                )
            if azimuth_delta:
                ax.azim = ax.azim + azimuth_delta  # type: ignore
            if elevation_delta:
                ax.elev = ax.elev + elevation_delta  # type: ignore
            if roll_delta:
                ax.roll = ax.roll + roll_delta  # type: ignore

        # update data
        points = np.array(points)
        scatter._offsets3d = (  # pylint: disable=protected-access
            points[..., 0].ravel(),
            points[..., 1].ravel(),
            points[..., 2].ravel(),
        )
        for idx_track, line in zip(line_indexes, lines):
            line.set_data_3d(points[..., idx_track, :3].T)

        return [scatter] + list(lines)

    @classmethod
    def animate(
        cls,
        frames: Union[Sequence[Sequence[Tuple[float, float, float]]], NDArray],
        line_indexes: Optional[Sequence[Sequence[int]]] = None,
        line_colors: Sequence[Union[Tuple[float, float, float], None]] = (),
        line_labels: Sequence[Union[str, None]] = (),
        scatter_color: Tuple[float, float, float] = (0, 0, 0),
        scatter_size: float = 2,
        title: Optional[str] = "{frame_number}",
        vertical_axis: Literal["x", "y", "z"] = "z",
        ticks_scale: Optional[float] = None,
        azimuth: float = 20,
        elevation: float = 15,
        roll: float = 0,
        azimuth_delta: float = 0,
        elevation_delta: float = 0,
        roll_delta: float = 0,
        invert_x: bool = False,
        invert_y: bool = False,
        invert_z: bool = False,
        show_grid: bool = True,
        show_axis: bool = True,
        figure_scale: Optional[float] = None,
        style: Literal["dark_background", "default"] = "default",
        layout: Literal["constrained", "compressed", "tight", "none"] = "none",
        interval: Union[float, int] = 37,
        repeat_delay: Union[float, int] = 100,
        blit: bool = True,
    ) -> FuncAnimation:
        """
        Animates the given frames representing 3D coordinates with 3D scatter plot and lines connecting those points.

        Args:
            frames (Union[Sequence[Sequence[Tuple[float, float, float]]], NDArray]): The frames to animate, represented as a sequence of collection of 3D coordinates.
            line_indexes (Optional[Sequence[Sequence[int]]]): The indexes of the points in a frame to connect in lines.  If not provided, connects the points in a cycle [0, 1, 2, ..., n-1, 0].
            line_colors (Sequence[Union[Tuple[float, float, float], None]]): The colors of the lines in RGB format normalized to [0.0, 1.0] range. If not provided, default to a gradient of blue to pink to blue.
            line_labels (Sequence[Union[str, None]]): The labels of the lines.
            scatter_color (Tuple[float, float, float]): The color of the scatter points in RGB format normalized to [0.0, 1.0] range. Default is black.
            title (Optional[str]): The title of the animation. Can include the placeholder "{frame_number}" to display the frame number. Defaults to "{frame_number}".
            vertical_axis (Literal["x", "y", "z"]): The vertical axis in the plot. Default is "z".
            ticks_scale (Optional[float]): The scale of the ticks. Defaults to the nearest power of 10 under the range in data.
            layout (Literal["constrained", "compressed", "tight", "none"]): The layout of the plot. Default is "none".
            interval (Union[float, int]): The interval between frames in milliseconds. Default is 37.
            repeat_delay (Union[float, int]): The delay between replays in milliseconds. Default is 100.
            blit (bool): Whether to use blitting for faster updates. Default is True.

        Returns:
            FuncAnimation: The animation object.
        """
        frames = np.array(frames)
        limits = np.stack([frames.min(axis=(0, 1))[:3], frames.max(axis=(0, 1))[:3]]).T

        if line_indexes is None:
            line_indexes = _indexes_to_connect(len(frames[0]))
        if not ticks_scale:
            ticks_scale = _nearest_scale(limits.ptp(1).min())

        fig, [ax] = cls.new_figure(
            *limits,
            vertical_axis=vertical_axis,
            figure_scale=figure_scale,
            style=style,
            layout=layout,
            subplots=(1, 1),
        )
        cls.initialize_Axes3D(
            ax,
            *limits,
            ticks_scale=ticks_scale,
            azimuth=azimuth,
            elevation=elevation,
            roll=roll,
            vertical_axis=vertical_axis,
            invert_x=invert_x,
            invert_y=invert_y,
            invert_z=invert_z,
            show_grid=show_grid,
            show_axis=show_axis,
        )

        scatter, lines = cls.placeholder_scatter_and_lines(
            ax,
            len(line_indexes),
            line_colors=line_colors,
            line_labels=line_labels,
            scatter_color=scatter_color,
            scatter_size=scatter_size,
        )

        if line_labels:
            ax.legend(loc="center right")
        if title:
            # TODO: why is -4 required here to show t_0 in the fist frame???
            ax.set_title(title.format(frame_number="[t_-4]"))

        animation = FuncAnimation(
            fig,
            partial(
                cls.set_frame_data,
                scatter=scatter,
                lines=lines,
                line_indexes=line_indexes,
                ax=ax,
                azimuth_delta=azimuth_delta,
                elevation_delta=elevation_delta,
                roll_delta=roll_delta,
            ),
            frames=frames,
            interval=interval,
            blit=blit,
            repeat_delay=repeat_delay,
        )

        # plt.close(fig)  # todo: handle trailing figure

        return animation

    @classmethod
    def frames_grid(
        cls,
        frames: Union[Sequence[Sequence[Tuple[float, float, float]]], NDArray],
        subplots: Tuple[int, int],
        line_indexes: Optional[Sequence[Sequence[int]]] = None,
        line_colors: Sequence[Union[Tuple[float, float, float], None]] = (),
        line_labels: Sequence[Union[str, None]] = (),
        scatter_color: Tuple[float, float, float] = (0, 0, 0),
        scatter_size: float = 2,
        title: Optional[str] = "{frame_number}",
        figure_title: Optional[str] = None,
        figure_title_font_size: float = 20,
        vertical_axis: Literal["x", "y", "z"] = "z",
        ticks_scale: Optional[float] = None,
        azimuth: float = 20,
        elevation: float = 15,
        roll: float = 0,
        azimuth_delta: float = 0,
        elevation_delta: float = 0,
        roll_delta: float = 0,
        invert_x: bool = False,
        invert_y: bool = False,
        invert_z: bool = False,
        show_grid: bool = True,
        show_axis: bool = True,
        figure_scale: Optional[float] = 4,
        style: Literal["dark_background", "default"] = "default",
        layout: Literal["constrained", "compressed", "tight", "none"] = "tight",
    ) -> Figure:
        """
        Generates a grid of frames with 3D scatter plots and lines connecting the points.

        Args:
            frames (Union[Sequence[Sequence[Tuple[float, float, float]]], NDArray]): The frames containing the 3D coordinates of the points.
            subplots (Tuple[int, int]): The number of rows and columns in the figure. Each cell is a 3D plot containing one frame.
            line_indexes (Optional[Sequence[Sequence[int]]]): The indexes of points to be connected with lines.
            line_colors (Sequence[Union[Tuple[float, float, float], None]]): The colors of the lines connecting the points. Color should be in RGB format and in range [0.0, 1.0].
            line_labels (Sequence[Union[str, None]]): The labels for the lines connecting the points.
            scatter_color (Tuple[float, float, float]): The color of the scatter points. Color should be in RGB format and in range [0.0, 1.0].
            scatter_size (float): The size of the scatter points.
            title (Optional[str]): The title of each subplot. Can include the placeholder "{frame_number}" to display the frame number.
            figure_title (Optional[str]): The title of the entire figure.
            figure_title_font_size (float): The font size of the figure title.
            vertical_axis (Literal["x", "y", "z"]): The vertical axis in the 3D plots.
            azimuth (float): The azimuth angle (rotation around the vertical axis) of the initial view in the plot. Value must be in degrees.
            elevation (float): The elevation angle (amount of rise from the horizontal plane) of the initial view in the plot.  Value must be in degrees.
            roll (float): The roll angle (rotation around the line of sight) of the initial view in the plot. Value must be in degrees.
            azimuth_delta (float): The change in azimuth angle for each subplot. Value must be in degrees.
            elevation_delta (float): The change in elevation angle for each subplot. Value must be in degrees.
            roll_delta (float): The change in roll angle for each subplot. Value must be in degrees.
            invert_x (bool): Whether to invert the x-axis.
            invert_y (bool): Whether to invert the y-axis.
            invert_z (bool): Whether to invert the z-axis.
            show_grid (bool): Whether to show the grid lines on the axes.
            show_axis (bool): Whether to show the axis lines.
            figure_scale (Optional[float]): The size of the entire figure.
            style (Literal["dark_background", "default"]): The color theme of the plot.
            layout (Literal["constrained", "compressed", "tight", "none"]): The spacing between the subplots.

        Returns:
            Figure: The generated matplotlib figure.
        """
        frames = np.array(frames)
        limits = np.stack([frames.min(axis=(0, 1))[:3], frames.max(axis=(0, 1))[:3]]).T
        indexes = np.linspace(0, len(frames) - 1, np.prod(subplots)).round().astype(int)

        if line_indexes is None:
            line_indexes = _indexes_to_connect(len(frames[0]))
        if not ticks_scale:
            ticks_scale = _nearest_scale(limits.ptp(1).min())

        fig, axes = cls.new_figure(
            *limits,
            vertical_axis=vertical_axis,
            figure_scale=figure_scale,
            style=style,
            layout=layout,
            subplots=subplots,
        )

        for i, ax in enumerate(axes):
            cls.initialize_Axes3D(
                ax,
                *limits,
                ticks_scale=ticks_scale,
                azimuth=azimuth + i * azimuth_delta,
                elevation=elevation + i * elevation_delta,
                roll=roll + i * roll_delta,
                vertical_axis=vertical_axis,
                invert_x=invert_x,
                invert_y=invert_y,
                invert_z=invert_z,
                show_grid=show_grid,
                show_axis=show_axis,
            )

            scatter, lines = cls.placeholder_scatter_and_lines(
                ax,
                len(line_indexes),
                line_colors=line_colors,
                line_labels=line_labels,
                scatter_color=scatter_color,
                scatter_size=scatter_size,
            )

            if title:
                ax.set_title(title.format(frame_number=f"[t_{indexes[i]}]"))

            cls.set_frame_data(
                frames[indexes[i]], scatter, lines, line_indexes=line_indexes
            )

        if line_labels:
            fig.legend(handles=axes[0].lines, loc="center right")

        if figure_title:
            fig.suptitle(figure_title, fontsize=figure_title_font_size)

        return fig


def _indexes_to_connect(n: int) -> List[List[int]]:
    return [[i, (i + 1) % n] for i in range(n)]


def _nearest_scale(x: float, base: int = 10) -> float:
    return float(base ** math.floor(math.log(x, base)))


def _reset_counter_in_animation_title(animation: FuncAnimation) -> None:
    if isinstance(animation, FuncAnimation):
        ax = animation._fig.gca()  # type: ignore  # pylint: disable=protected-access
        ax.set_title(re.sub(r"\[t_-?\d+\]", "[t_-2]", ax.get_title()))


# todo: class Plotly3D:
# todo: class Avatar3D: # uses open3d and custom avatar



from typing import Dict, List, Optional, Sequence, Tuple, Type, Union

import numpy as np
import torch
from numpy.typing import NDArray
from torch import Tensor

__all__ = [
    "ArrayOps",
    "linear_interpolation",
    "adjust_vector_angle",
    "align_vectors",
]


class ArrayOps:
    @staticmethod
    def floor(
        array: Union[NDArray, Tensor, Sequence[Union[float, int]], float, int]
    ) -> Union[NDArray, Tensor]:
        if isinstance(array, (np.ndarray, list, tuple, float, int)):
            return np.floor(array)

        if isinstance(array, Tensor):
            return array.floor()

        raise TypeError(f"Invalid type for flooring: {type(array)}")

    @staticmethod
    def ceil(
        array: Union[NDArray, Tensor, Sequence[Union[float, int]], float, int]
    ) -> Union[NDArray, Tensor]:
        if isinstance(array, (np.ndarray, list, tuple, float, int)):
            return np.ceil(array)

        if isinstance(array, Tensor):
            return array.ceil()

        raise TypeError(f"Invalid type for ceiling: {type(array)}")

    @staticmethod
    def take(
        array: Union[NDArray, Tensor, List],
        index: Union[NDArray, Tensor, List, int],
        dim: int = 0,
    ) -> Union[NDArray, Tensor]:
        if isinstance(array, (np.ndarray, list, tuple)):
            if not isinstance(index, np.ndarray):
                index = np.array(index)
            return np.take(array, index.astype(int), axis=dim)

        if isinstance(array, Tensor):
            if not isinstance(index, Tensor):
                index = torch.tensor(index)
            return array.index_select(dim, index.type(torch.long))

        raise TypeError(f"Invalid type for taking: {type(array)}")

    @staticmethod
    def cast(
        x: Union[NDArray, Tensor, Sequence[Union[float, int]]],
        data_type: Type[Union[np.ndarray, Tensor]],
        _dtype: Optional[Union[Type[torch.dtype], Type[np.dtype], Type]] = None,
    ) -> Union[NDArray, Tensor]:

        if data_type == np.ndarray:
            return np.array(x, dtype=_dtype)

        if data_type == Tensor:
            type_map: Dict = {int: torch.int64, float: torch.float64}
            _dtype = type_map.get(_dtype, _dtype)
            if isinstance(x, np.ndarray):
                x = torch.from_numpy(x)
            if isinstance(x, Tensor):
                return x.to(dtype=_dtype)  # type: ignore
            return torch.tensor(x, dtype=_dtype)  # type: ignore

        raise ValueError(f"Invalid data_type for array casting: {data_type}")

    @staticmethod
    def norm(
        x: Union[NDArray, Tensor, Sequence[Union[float, int]]],
        dim: Optional[int] = None,
        keepdim=False,
    ) -> Union[NDArray, Tensor]:
        """
        Compute the norm of a given array or tensor along a specified dimension.

        Args:
            x (Union[NDArray, Tensor, Sequence[Union[float, int]]]): The input array or tensor.
            dim (Optional[int]): The dimension along which to compute the norm. If None, the norm is computed over the entire array or tensor. Default is None.
            keepdim (bool): Whether to keep the dimension of the input array or tensor after computing the norm. Default is False.

        Returns:
            Union[NDArray, Tensor]: The norm of the input array or tensor.

        Raises:
            TypeError: If the input type is not supported.
        """

        if isinstance(x, (np.ndarray, list, tuple)):
            return np.linalg.norm(x, axis=dim, keepdims=keepdim)  # type: ignore

        if isinstance(x, Tensor):
            return x.norm(dim=dim, keepdim=keepdim)

        raise TypeError(f"Invalid type for norm: {type(x)}")

    @staticmethod
    def svd(
        x: Union[NDArray, Tensor, Sequence[Sequence[Union[float, int]]]]
    ) -> Tuple[Union[NDArray, Tensor], Union[NDArray, Tensor], Union[NDArray, Tensor]]:
        """
        Compute the singular value decomposition of a given array or tensor.

        Args:
            x (Union[NDArray, Tensor, Sequence[Sequence[Union[float, int]]]]): The input array or tensor.

        Returns:
            Tuple[Union[NDArray, Tensor], Union[NDArray, Tensor], Union[NDArray, Tensor]]: The (Rotation, coordinate scaling, reflection) matrices of the input array or tensor.

        Raises:
            TypeError: If the input type is not supported.
        """

        if isinstance(x, (np.ndarray, list, tuple)):
            return np.linalg.svd(x)

        if isinstance(x, Tensor):
            U, S, V = x.svd()
            return U, S, V.T

        raise TypeError(f"Invalid type for svd: {type(x)}")

    @staticmethod
    def top_k(
        x: Union[NDArray, Tensor, Sequence[Union[float, int]]],
        k: int,
        dim: int = -1,
        largest=True,
    ) -> Tuple[Union[NDArray, Tensor], Union[NDArray, Tensor]]:
        """
        Compute the top k values and their indices along a specified dimension of a given array or tensor.

        Args:
            x (Union[NDArray, Tensor, Sequence[Union[float, int]]]): The input array or tensor.
            k (int): The number of top values to return.
            dim (int, optional): The dimension along which to compute the top k values. Default is -1.
            largest (bool, optional): Whether to return the largest or smallest k values. Default is True.

        Returns:
            Tuple[Union[NDArray, Tensor], Union[NDArray, Tensor]]: The top k values and their indices along the specified dimension.

        Raises:
            TypeError: If the input type is not supported.
        """

        if isinstance(x, np.ndarray):
            indices = (
                np.argpartition(x, -k, axis=dim)[-k:]
                if largest
                else np.argpartition(x, k, axis=dim)[:k]
            )
            return x.take(indices, axis=dim), indices
        if isinstance(x, Tensor):
            values, indices = x.topk(k, dim=dim, largest=largest, sorted=True)
            return values, indices

        raise TypeError(f"Invalid type for topk: {type(x)}")

    @staticmethod
    def concatenate(
        arrays: Union[Sequence[NDArray], Sequence[Tensor]], dim: int = 0
    ) -> Union[NDArray, Tensor]:
        """
        Concatenate a sequence of arrays or tensors along a specified dimension.

        Args:
            arrays (Union[Sequence[NDArray], Sequence[Tensor]]): The sequence of arrays or tensors to concatenate.
            dim (int, optional): The dimension along which to concatenate the arrays or tensors. Default is 0.

        Returns:
            Union[NDArray, Tensor]: The concatenated array or tensor.

        Raises:
            TypeError: If the input type is not supported.
        """

        if isinstance(arrays[0], np.ndarray):
            return np.concatenate(arrays, axis=dim)
        if isinstance(arrays[0], Tensor):
            return torch.concatenate(
                [(x if isinstance(x, Tensor) else torch.tensor(x)) for x in arrays],
                dim=dim,
            )

        raise TypeError(f"Invalid type for concatenation: {type(arrays[0])}")

    @staticmethod
    def abs(x: Union[NDArray, Tensor]) -> Union[NDArray, Tensor]:
        """
        Compute the absolute value of a given array or tensor.

        Args:
            x (Union[NDArray, Tensor]): The input array or tensor.

        Returns:
            Union[NDArray, Tensor]: The absolute value of the input array or tensor.

        Raises:
            TypeError: If the input type is not supported.
        """

        if isinstance(x, np.ndarray):
            return np.abs(x)
        if isinstance(x, Tensor):
            return x.abs()

        raise TypeError(f"Invalid type for absolute value: {type(x)}")

    @staticmethod
    def copy(x: Union[NDArray, Tensor]) -> Union[NDArray, Tensor]:
        """
        Create a copy of a given array or tensor.

        Args:
            x (Union[NDArray, Tensor]): The input array or tensor.

        Returns:
            Union[NDArray, Tensor]: A deep copy of the input array or tensor.

        Raises:
            TypeError: If the input type is not supported.
        """

        if isinstance(x, np.ndarray):
            return np.copy(x)
        if isinstance(x, Tensor):
            return torch.clone(x)

        raise TypeError(f"Invalid type for copying array: {type(x)}")


def linear_interpolation(
    array: Union[NDArray[np.number], Tensor, Sequence],
    new_x: Union[Sequence[Union[int, float]], NDArray[np.number], Tensor],
    old_x: Union[Sequence[Union[int, float]], NDArray[np.number], Tensor, None] = None,
    dim: int = 0,
) -> Union[NDArray, Tensor]:
    """
    Perform linear interpolation on a multidimensional array or tensor along a dimension.

    This function essentially connects all consecutive values in a multidimensional array with
    straight lines along a specified dimension, so that intermediate values can be calculated.
    It takes the input array, a set of new indexes or alternatively new & old coordinate values,
    and a dimension along which to perform interpolation.

    Parameters:
        array (NDArray[np.number] | Tensor | List): The input array or tensor to interpolate.
        new_x (Sequence[int | float] | NDArray[np.number] | Tensor): The new index values or coordinate values at which to calculate the intermediate values from `array`. Must be 1D. Order of values does not matter. if `old_x` is not provided, these values are relative to the index of the data in `array` i.e. [0, 1, 2, ...] and negative indexes are allowed. If `old_x` is provided, all `new_x` values must be within it's bounds.
        old_x (Sequence[int | float] | NDArray[np.number] | Tensor | None, optional): The old *coordinate* values corresponding to the data in `array` along the `dim`. Must be 1D and strictly sorted. Can contain negative numbers. If None, method assumes it to be a linear sequence starting at 0 and growing with step +1 i.e. `[0, 1, 2, ...]` like the index of `array`.
        dim (int, optional): The dimension along which to perform interpolation. Default is 0.

    Returns:
        NDArray | Tensor: The result of linear interpolation along the specified dimension.

    Raises:
        ValueError: If `new_x` or `old_x` is not 1 dimensional.

    Examples:

        .. code-block: python

            data = np.array([1, 2, 3, 5])
            new_indexes = np.array([1.5, 0.5, 2.5])
            interpolated_data = linear_interpolation(data, new_indexes)
            print(interpolated_data)
            # array([2.5, 1.5, 4. ])

            old_x = [0, 4, 4.5, 5]
            new_x = [0, 1, 2, 2.5, 3, 4, 5]
            interpolated_data = linear_interpolation(data, new_x, old_x=old_x)
            print(interpolated_data)
            # array([1.   , 1.25 , 1.5  , 1.625, 1.75 , 2.   , 5.   ])

    Note:
        This function supports both NumPy arrays and PyTorch tensors as input and preserves gradient.
    """
    if not isinstance(array, (np.ndarray, Tensor)):
        array = np.array(array)

    indexes, dim = __validate_lin_interp_args(array, new_x, old_x, dim)

    # interpolate (Magic!)
    floored_indexes = ArrayOps.floor(indexes)
    ceiled_indexes = ArrayOps.ceil(indexes)
    fraction = (indexes - floored_indexes).reshape(
        [len(indexes) if d == dim else 1 for d in range(array.ndim)]
    )

    interpolated = ArrayOps.take(array, floored_indexes, dim) * (1 - fraction)
    interpolated = interpolated + ArrayOps.take(array, ceiled_indexes, dim) * fraction

    return interpolated


def __validate_lin_interp_args(array: Union[NDArray, Tensor], new_x, old_x, dim: int):
    if dim < -1 * array.ndim or dim >= array.ndim:
        raise ValueError(f"Invalid dim: {dim} use -{array.ndim} <= d <= {array.ndim-1}")
    if dim < 0:
        dim = dim + array.ndim

    if old_x is not None:
        if len(old_x) != array.shape[dim]:
            raise ValueError(f"Invalid old_x: ({len(old_x)=}) != ({array.shape[dim]=})")

        old_x, new_x = np.array(old_x), np.array(new_x)
        if (out_of_bounds := (new_x < old_x.min()) | (new_x > old_x.max())).any():
            raise ValueError(
                f"Invalid values in new_x: {new_x[out_of_bounds].tolist()}. "
                f"Must be in range {old_x.min()} <= x <= {old_x.max()}."
            )

        differences = old_x[:-1] - old_x[1:]
        if not ((is_descending := (differences > 0).all()) or (differences < 0).all()):
            raise ValueError("Provide sorted `old_x` (otherwise rearrange the array)")

        indexes = (
            np.interp(new_x, old_x[::-1], np.arange(len(old_x))[::-1])
            if is_descending
            else np.interp(new_x, old_x, np.arange(len(old_x)))
        )

    else:
        if not isinstance(new_x, (np.ndarray, Tensor)):
            new_x = np.array(new_x)
        if (
            out_of_bounds := (new_x < -array.shape[dim]) | (new_x >= array.shape[dim])
        ).any():
            raise ValueError(
                f"Invalid new indexes: {new_x[out_of_bounds.tolist()].tolist()}. "
                f"Must be in range {-array.shape[dim]} <= x <= {array.shape[dim]-1}."
            )
        indexes = new_x

    indexes = ArrayOps.cast(indexes, type(array))
    if indexes.ndim != 1:
        raise ValueError(f"Invalid indexes shape: {indexes.shape}. Must be 1D.")

    return indexes, dim


def adjust_vector_angle(
    vector_1: Union[NDArray, Tensor, Sequence[float]],
    vector_2: Union[NDArray, Tensor, Sequence[float]],
    scaling_factor: float,
    post_normalize: bool = False,
) -> Tuple[Union[NDArray, Tensor], Union[NDArray, Tensor]]:
    """Move a pair of vectors away or towards each other in the same plane.

    Converge or Diverge a pair of vectors by increasing or decreasing their distance from each other.
    The norm or the length of the vectors is preserved.

    Args:
        vector_1 (NDArray | Tensor): A 1D array of size n representing a word in an n dimensional vector space.
        vector_2 (NDArray | Tensor): A 1D array of size n representing another word in an n dimensional vector space.
        scaling_factor (float): The scaling factor by which the vector difference should be enhanced or diminished. The fraction of distance between the vectors where new vector should land. (sf > 1 diverges the two vectors. sf = 1 leaves the two vectors unchanged. 0.5 < sf < 1 converges the two vectors. sf = 0.5 makes the two vectors equal to their mean. sf = 0 swaps the two vectors. sf < 0.5 move the vectors away from their mean but in opposite direction.)
        post_normalize (bool, optional): Make the magnitude of both output vectors equal to 1 after they have been rotated. Defaults to False.

    Returns:
        Tuple[NDArray | Tensor, NDArray | Tensor]: moved vectors.
    """

    # TODO: make it work for batches of vectors or nd-arrays
    # TODO: Handle NaNs

    if not isinstance(vector_1, (np.ndarray, Tensor)):
        vector_1 = np.array(vector_1)
    if not isinstance(vector_2, (np.ndarray, Tensor)):
        vector_2 = np.array(vector_2)

    v1_norm = ArrayOps.norm(vector_1, dim=None)
    v2_norm = ArrayOps.norm(vector_2, dim=None)

    # make the magnitude of both vectors = 1
    vector_1 = vector_1 / (v1_norm or 1)
    vector_2 = vector_2 / (v2_norm or 1)

    # figure out the dimension of divergence
    v1_minus_v2 = vector_1 - vector_2

    # move each vector away or towards the other by equal amount
    new_v1 = vector_2 + scaling_factor * v1_minus_v2
    new_v2 = vector_1 - scaling_factor * v1_minus_v2

    # make the magnitude of the new vectors = 1
    new_v1 = new_v1 / (ArrayOps.norm(new_v1) or 1)
    new_v2 = new_v2 / (ArrayOps.norm(new_v2) or 1)

    if not post_normalize:
        # restore original magnitudes
        new_v1 = new_v1 * v1_norm
        new_v2 = new_v2 * v2_norm

    # sf > 1 diverges the two vectors
    # new_v1 = v2 + 2.00 * (v1 - v2) = 2 * v1 - v2     # more v1, less v2.
    # new_v2 = v1 - 2.00 * (v1 - v2) = 2 * v2 - v1     # more v2, less v1.

    # sf = 1 leaves the two vectors unchanged
    # new_v1 = v2 + 1.00 * (v1 - v2) = v1
    # new_v2 = v1 - 1.00 * (v1 - v2) = v2

    # 0.5 < sf < 1 converges the two vectors
    # new_v1 = v2 + 0.75 * (v1 - v2) = 0.75 * v1 + 0.25 * v2    # weighted average
    # new_v1 = v1 - 0.75 * (v1 - v2) = 0.75 * v2 + 0.25 * v1    # weighted average

    # sf = 0.5 makes the two vectors equal
    # new_v1 = v2 + 0.50 * (v1 - v2) = 0.5 * v1 + 0.5 * v2   # mean
    # new_v1 = v1 - 0.50 * (v1 - v2) = 0.5 * v2 + 0.5 * v1   # mean

    # sf = 0. swaps the two vectors
    # new_v1 = v2 + 0.00 * (v1 - v2) = v2
    # new_v2 = v1 + 0.00 * (v1 - v2) = v1

    # sf < 0.5 move the vectors away from their mean but in opposite direction
    # new_v1 = v2 + (-1) * (v1 - v2) = 2 * v2 - v1    # more v2, less v1.
    # new_v2 = v1 - (-1) * (v1 - v2) = 2 * v1 - v2    # more v1, less v2.

    return new_v1, new_v2


def align_vectors(
    source_matrix: Union[NDArray, Tensor],
    target_matrix: Union[NDArray, Tensor],
    pre_normalize: bool = True,
) -> Union[NDArray, Tensor]:
    """
    Align the source matrix to the target matrix using the orthogonal transformation.

    Args:
        source_matrix (NDArray | Tensor): A 2D array of shape (dictionary_length, embedding_dimension) containing word vectors from source model (or language).
        target_matrix (NDArray | Tensor): A 2D array of shape (dictionary_length, embedding_dimension) containing word vectors from target model (or language).
        normalize_vectors (bool, optional): Whether to normalize the training vectors before SVD. Defaults to True.

    Returns:
        NDArray | Tensor: An orthogonal transformation which aligns the source language to the target language.

    Note:
        This function supports both NumPy arrays and PyTorch tensors as input.
        (Based on: https://github.com/babylonhealth/fastText_multilingual)
    """

    if not isinstance(source_matrix, (np.ndarray, Tensor)):
        source_matrix = np.array(source_matrix)
    if not isinstance(target_matrix, (np.ndarray, Tensor)):
        target_matrix = np.array(target_matrix)

    # optionally normalize the training vectors
    if pre_normalize:
        src_norm = ArrayOps.norm(source_matrix, dim=1, keepdim=True)
        src_norm[src_norm == 0] = 1
        source_matrix = source_matrix / src_norm

        tgt_norm = ArrayOps.norm(target_matrix, dim=1, keepdim=True)
        tgt_norm[tgt_norm == 0] = 1
        target_matrix = target_matrix / tgt_norm

    # perform the SVD
    product = source_matrix.T @ target_matrix
    U, _, V = ArrayOps.svd(product)

    # return orthogonal transformation which aligns source language to the target
    return U @ V



"""Centralized String Constants for Sign Language Translator Package

This module defines enumerations for various string constants used throughout the package.

Enumerations:
    Countries (Enum): Enumerates supported countries with their short codes.
    Organizations (Enum): Enumerates supported organizations with their short codes.
    SignCollections (Enum): Enumerates sign collections with their names.
    TextLanguages (Enum): Enumerates supported text languages with their short codes.
    SignLanguages (Enum): Enumerates supported sign languages with their names.
    SignFormats (Enum): Enumerates supported video feature models.
    ModelCodes (Enum): Enumerates model codes for different models.

Functions:
    normalize_short_code (function): Normalizes a provided short code to a recognized standard form.
"""

from __future__ import annotations

from enum import Enum
from typing import Union

from sign_language_translator.utils.utils import (
    PrintableEnumMeta,
    search_in_values_to_retrieve_key,
)

__all__ = [
    "Countries",
    "Organizations",
    "SignCollections",
    "TextLanguages",
    "SignLanguages",
    "SignFormats",
    "ModelCodes",
    "SignEmbeddingModels",
    "normalize_short_code",
]


class Countries(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of countries with their corresponding short codes.

    Attributes:
        - PAKISTAN (str): Short code for Pakistan.
        ...
    """

    PAKISTAN = "pk"
    # INDIA = "in"
    # USA = "us"


class Organizations(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of organizations with their corresponding short codes.

    Attributes:
        - HFAD (str): Short code for HFAD (e.g., an organization for deaf in Pakistan).
        ...
    """

    HFAD = "hfad"
    # NISE = "nise"
    # FESF = "fesf"


class SignCollections(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of sign collections with their corresponding short codes.

    Attributes:
        - PK_HFAD_1 (str): Short code for the first sign dictionary from HamzaFoundationAcademyDeaf, Pakistan.
        - PK_HFAD_2 (str): Short code for the second sign collection from HamzaFoundationAcademyDeaf, Pakistan.
        ...
    """

    PK_HFAD_1 = f"{Countries.PAKISTAN}-{Organizations.HFAD}-1"
    """Short code for the first sign dictionary from HamzaFoundationAcademyDeaf, Pakistan. (788 videos)"""

    PK_HFAD_2 = f"{Countries.PAKISTAN}-{Organizations.HFAD}-2"
    """Short code for the second sign dictionary from HamzaFoundationAcademyDeaf, Pakistan."""

    # PK_FESF_1 = f"{Countries.PAKISTAN}-{Organizations.FESF}-1"
    # PK_NISE_1 = f"{Countries.PAKISTAN}-{Organizations.NISE}-1"


class TextLanguages(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of text languages with their corresponding short codes.

    Attributes:
        - URDU (str): Short code for the Urdu language.
        - HINDI (str): Short code for the Hindi language.
        ...
    """

    URDU = "ur"
    # ENGLISH = "en"
    HINDI = "hi"


class SignLanguages(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of sign languages with their corresponding short codes.

    Attributes:
        - PAKISTAN_SIGN_LANGUAGE (str): Short code for the Pakistan Sign Language.
        ...
    """

    PAKISTAN_SIGN_LANGUAGE = "pakistan-sign-language"


class SignFormats(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of available sign formats with their corresponding short codes.
    For example, sign language can be a sequence of frames (video) or a sequence of pose vectors (landmarks) etc.

    Attributes:
        - VIDEO (str): Short code for raw video.
        - LANDMARKS (str): Short code for body landmarks.
        ...
    """

    # raw video
    VIDEO = "video"
    """Sequence of RGB frames (num_frames, height, width, num_channels)"""

    # Body Landmarks
    LANDMARKS = "landmarks"
    """Sequence of points on the body (n_frames, n_points, n_coordinates)"""

    # MEDIAPIPE_LANDMARKS = "mediapipe-landmarks"
    # """3d world coordinates (x,y,z,[visibility, presence]) and/or 2d image coordinates (x,y,depth,[visibility, presence])"""

    # Body Mesh Grid

    # Image Segmentation

    # Motion Vectors


class ModelCodes(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration of model codes with their corresponding short codes.

    Attributes:
        - CONCATENATIVE_SYNTHESIS (str): Short code for the rule-based text-to-sign translation model.

        - NGRAM_LM_UNIGRAM_NAMES (str): Short code for ngram model trained with window size 1 on en/ur person names data.
        - NGRAM_LM_BIGRAM_NAMES (str): Short code for ngram model trained with window size 2 on en/ur person names data.
        - NGRAM_LM_TRIGRAM_NAMES (str): Short code for ngram model trained with window size 3 on en/ur person names data.
        - MIXER_LM_NGRAM_URDU (str): Short code for a mix of ngram models trained on urdu words of window size 1 to 6.
        - TRANSFORMER_LM_UR_SUPPORTED (str): Short code for a transformer-based language model trained on ur supported tokens.

        - MEDIAPIPE_POSE_V2_HAND_V1 (str): Short code for a video embedding model which uses MediaPipe pose_landmarker_heavy & hand_landmarker.
        - MEDIAPIPE_POSE_V1_HAND_V1 (str): Short code for a video embedding model which uses MediaPipe pose_landmarker_full & hand_landmarker.
        - MEDIAPIPE_POSE_V0_HAND_V1 (str): Short code for a video embedding model which uses MediaPipe pose_landmarker_lite & hand_landmarker.

        - LOOKUP_UR_FASTTEXT_CC (str): Short code for a text embedding model which uses a lookup table to embed Urdu tokens using fastText embeddings trained on Common Crawl.
    """

    # text-to-sign translation
    CONCATENATIVE_SYNTHESIS = "concatenative-synthesis"
    """The core rule-based text to sign translation model that joins sign clips for each word in a text sentence."""
    # LANDMARK_GAN = "landmark-gan"

    # sign-to-text translation
    # GESTURE = "gesture"

    # language-models
    NGRAM_LM_UNIGRAM_NAMES = "names-stat-lm-w1.json"
    NGRAM_LM_BIGRAM_NAMES = "names-stat-lm-w2.json"
    NGRAM_LM_TRIGRAM_NAMES = "names-stat-lm-w3.json"
    MIXER_LM_NGRAM_URDU = "ur-supported-token-unambiguous-mixed-ngram-w1-w6-lm.pkl"
    """Simple hash table based n-gram language model with context size of 1-6 that generates unambiguous Urdu tokens."""
    TRANSFORMER_LM_UR_SUPPORTED = "tlm_14.0M.pt"

    # video-embedding-models
    MEDIAPIPE_POSE_V2_HAND_V1 = "mediapipe-pose-2-hand-1"
    """Short code for the video embedding model which uses MediaPipe pose_landmarker_heavy & hand_landmarker to generate (33 pose + 2 * 21 hand) world & 75 image landmarks (x, y, z, visibility, presence) for each frame of the video."""
    MEDIAPIPE_POSE_V1_HAND_V1 = "mediapipe-pose-1-hand-1"
    """Short code for the video embedding model which uses MediaPipe pose_landmarker_full  & hand_landmarker to generate (33 pose + 2 * 21 hand) world & 75 image landmarks (x, y, z, visibility, presence) for each frame of the video."""
    MEDIAPIPE_POSE_V0_HAND_V1 = "mediapipe-pose-0-hand-1"
    """Short code for the video embedding model which uses MediaPipe pose_landmarker_lite  & hand_landmarker to generate (33 pose + 2 * 21 hand) world & 75 image landmarks (x, y, z, visibility, presence) for each frame of the video."""

    # text-embedding-models
    LOOKUP_UR_FASTTEXT_CC = "lookup-ur-fasttext-cc.pt"
    """Short code for the text embedding model which uses a lookup table to embed Urdu tokens using fastText embeddings trained on Common Crawl."""


class SignEmbeddingModels(Enum, metaclass=PrintableEnumMeta):
    """The Names of video embedding models that *have been* used to embed sign language videos in the available datasets.

    Attributes:
        MEDIAPIPE_WORLD (str): Short code for the video embedding model which uses MediaPipe pose_landmarker_heavy & hand_landmarker to generate (33 pose + 2 * 21 hand) world landmarks (x, y, z, visibility, presence) for each frame of the video. World landmarks are 3D coordinates in meters with origin at the center of the hips for pose_landmarker and at the center of each hand for hand_landmarker model.
        MEDIAPIPE_IMAGE (str): Short code for the video embedding model which uses MediaPipe pose_landmarker_heavy & hand_landmarker to generate (33 pose + 2 * 21 hand) image landmarks (x, y, z, visibility, presence) for each frame of the video. Image landmarks are 2D coordinates as fraction of the frame width or height with origin at the top-left corner of the frame and z value is the depth from the camera.
    """

    MEDIAPIPE_WORLD = "mediapipe-world"
    MEDIAPIPE_IMAGE = "mediapipe-image"

    # TODO: rename `Models` in class name to avoid confusion with `ModelCodes`


class ModelCodeGroups(Enum, metaclass=PrintableEnumMeta):
    """
    Enumeration class for grouping supported model codes, making it easier to filter various models.

    Attributes:
        ALL_LANGUAGE_MODELS (set): Set of model codes for all language models.
        ALL_NGRAM_LANGUAGE_MODELS (set): Set of model codes for all n-gram language models.
        ALL_TRANSFORMER_LANGUAGE_MODELS (set): Set of model codes for all transformer-based language models.
        ALL_MIXER_LANGUAGE_MODELS (set): Set of model codes for all mixer-based language models.

        ALL_VIDEO_EMBEDDING_MODELS (set): Set of model codes for all video embedding models.
        ALL_MEDIAPIPE_EMBEDDING_MODELS (set): Set of model codes for all MediaPipe-based video embedding models.

        ALL_TEXT_EMBEDDING_MODELS (set): Set of model codes for all text embedding models.
        ALL_VECTOR_LOOKUP_MODELS (set): Set of model codes for all vector lookup-based text embedding models.
    """

    # language models
    ALL_NGRAM_LANGUAGE_MODELS = {
        ModelCodes.NGRAM_LM_UNIGRAM_NAMES.value,
        ModelCodes.NGRAM_LM_BIGRAM_NAMES.value,
        ModelCodes.NGRAM_LM_TRIGRAM_NAMES.value,
    }
    ALL_TRANSFORMER_LANGUAGE_MODELS = {
        ModelCodes.TRANSFORMER_LM_UR_SUPPORTED.value,
    }
    ALL_MIXER_LANGUAGE_MODELS = {
        ModelCodes.MIXER_LM_NGRAM_URDU.value,
    }
    ALL_LANGUAGE_MODELS = (
        ALL_NGRAM_LANGUAGE_MODELS
        | ALL_TRANSFORMER_LANGUAGE_MODELS  # type: ignore
        | ALL_MIXER_LANGUAGE_MODELS
    )

    # video embedding models
    ALL_MEDIAPIPE_EMBEDDING_MODELS = {
        ModelCodes.MEDIAPIPE_POSE_V2_HAND_V1.value,
        ModelCodes.MEDIAPIPE_POSE_V1_HAND_V1.value,
        ModelCodes.MEDIAPIPE_POSE_V0_HAND_V1.value,
    }
    ALL_VIDEO_EMBEDDING_MODELS = ALL_MEDIAPIPE_EMBEDDING_MODELS

    # text embedding models
    ALL_VECTOR_LOOKUP_MODELS = {
        ModelCodes.LOOKUP_UR_FASTTEXT_CC.value,
    }
    ALL_TEXT_EMBEDDING_MODELS = ALL_VECTOR_LOOKUP_MODELS


# TODO: move the mapping list outside of the function. maybe convert it to a class variable.
def normalize_short_code(short_code: Union[str, Enum]) -> str:
    """
    Normalize the provided short code to a standard form that is recognized package wide.

    Args:
        short_code (str): The short code to be normalized.

    Returns:
        str: The normalized short code.
    """

    if isinstance(short_code, Enum):
        short_code = str(short_code.value)

    normalized_to_codes = {
        ModelCodes.CONCATENATIVE_SYNTHESIS.value: {
            "rule-based",
            "concatenative",
            "concatenativesynthesis",
            "concatenative-synthesis",
            "concatenative_synthesis",
            "concat-synth",
        },
        TextLanguages.URDU.value: {
            "urdu",
            "urd",
            "ur",
        },
        TextLanguages.HINDI.value: {
            "hindi",
            "hin",
            "hi",
        },
        SignLanguages.PAKISTAN_SIGN_LANGUAGE.value: {
            "psl",
            "pk-sl",
            "pakistan-sign-language",
            "pakistan-sl",
            "pakistansignlanguage",
            "pakistan_sign_language",
        },
        SignFormats.VIDEO.value: {
            "vid",
            "videos",
            "vids",
        },
        SignFormats.LANDMARKS.value: {
            "landmark",
            "lmrks",
            "lmrk",
        },
        ModelCodes.NGRAM_LM_UNIGRAM_NAMES.value: {
            "unigram-names",
        },
        ModelCodes.NGRAM_LM_BIGRAM_NAMES.value: {
            "bigram-names",
        },
        ModelCodes.NGRAM_LM_TRIGRAM_NAMES.value: {
            "trigram-names",
        },
        ModelCodes.MIXER_LM_NGRAM_URDU.value: {
            "urdu-mixed-ngram",
            "urdu-supported-ngram",
            "ur-mixed-ngram",
            "ur-supported-ngram",
        },
        ModelCodes.TRANSFORMER_LM_UR_SUPPORTED.value: {
            "ur-supported-gpt",
            "urdu-supported-gpt",
        },
        ModelCodes.MEDIAPIPE_POSE_V2_HAND_V1.value: {
            "mediapipe",
        },
        SignEmbeddingModels.MEDIAPIPE_IMAGE.value: {
            "landmark-mediapipe-image",
            "landmarks-mediapipe-image",
            "mediapipe-image-landmarks",
            "mediapipe-pose-2-hand-1-image",
            "mediapipe-pose-2-hands-1-image",
            "landmarks-mediapipe-pose-2-hand-1-image",
            "landmark-mediapipe-pose-2-hand-1-image",
            "landmarks-mediapipe-pose-2-hands-1-image",
            "landmark-mediapipe-pose-2-hands-1-image",
        },
        SignEmbeddingModels.MEDIAPIPE_WORLD.value: {
            "landmark-mediapipe-world",
            "landmarks-mediapipe-world",
            "mediapipe-world-landmarks",
            "mediapipe-pose-2-hand-1-world",
            "mediapipe-pose-2-hands-1-world",
            "landmarks-mediapipe-pose-2-hand-1-world",
            "landmark-mediapipe-pose-2-hand-1-world",
            "landmarks-mediapipe-pose-2-hands-1-world",
            "landmark-mediapipe-pose-2-hands-1-world",
        },
        ModelCodes.LOOKUP_UR_FASTTEXT_CC.value: {
            "lookup-ur-ft-cc",
            "lookup-ur-fasttext-cc",
            "ur-lookup-fasttext-cc",
            "ur-lookup-ftcc",
            "ur-lookup-ft-cc",
            "urdu-lookup-fasttext-cc",
            "urdu-lookup-ftcc",
            "urdu-lookup-ft-cc",
        },
    }
    normalized_to_codes = {
        k: v.union(
            {
                k,
                k.replace("-", "_"),
                k.replace("_", "-"),
                *map(lambda x: x.replace("-", "_"), v),
            }
        )
        for k, v in normalized_to_codes.items()
    }

    normalized = search_in_values_to_retrieve_key(short_code, normalized_to_codes)

    return normalized or short_code



"""
sign_language_translator
========================

Code: https://github.com/sign-language-translator/sign-language-translator
Help: https://slt.readthedocs.io
Demo: https://huggingface.co/sltAI

This project is an effort to bridge the communication gap between the hearing and the hearing-impaired community using Artificial Intelligence.
The goal is to provide a user friendly API to novel Sign Language Translation solutions that can easily adapt to any regional sign language.

Usage
-----

.. code-block:: python

    import sign_language_translator as slt

    # download dataset or models (if you need them for personal use)
    # (by default, resources are auto-downloaded within the install directory)
    # slt.Assets.set_root_dir("path/to/folder")  # Helps preventing duplication across environments or using cloud synced data
    # slt.Assets.download(r".*.json")  # downloads into asset_dir
    # print(slt.Assets.FILE_TO_URL.keys())  # All downloadable resources

    print("All available models:")
    print(list(slt.ModelCodes))  # slt.ModelCodeGroups
    # print(list(slt.TextLanguageCodes))
    # print(list(slt.SignLanguageCodes))
    # print(list(slt.SignFormatCodes))

    # -------------------------- TRANSLATE: text to sign --------------------------

    # The core model of the project (rule-based text-to-sign translator)
    # which enables us to generate synthetic training datasets
    model = slt.models.ConcatenativeSynthesis(
    text_language="urdu", sign_language="pk-sl", sign_format="video" )

    text = "یہ بہت اچھا ہے۔" # "This very good is."
    sign = model.translate(text) # tokenize, map, download & concatenate
    sign.show()


    model.text_language = slt.TextLanguageCodes.HINDI     # slt.languages.text.English()
    model.sign_format = slt.SignFormatCodes.LANDMARKS
    model.sign_embedding_model = "mediapipe-world"

    sign_2 = model.translate("कैसे हैं आप?") # "How are you?"
    sign_2.save("how-are-you.csv", overwrite=True)
    sign_2.save_animation("how-are-you.gif", overwrite=True)

    # -------------------------- TRANSLATE: sign to text --------------------------

    # sign = slt.Video("path/to/video.mp4")
    sign = slt.Video.load_asset("pk-hfad-1_آپ-کا-نام-کیا(what)-ہے")  # your name what is? (auto-downloaded)
    sign.show_frames_grid()

    # Extract Pose Vector for feature reduction
    embedding_model = slt.models.MediaPipeLandmarksModel()      # pip install "sign_language_translator[mediapipe]"  # (or [all])
    embedding = embedding_model.embed(sign.iter_frames())

    slt.Landmarks(embedding.reshape((-1, 75, 5)),
                connections="mediapipe-world").show()

    # # Load sign-to-text model (pytorch) (COMING SOON!)
    # translation_model = slt.get_model(slt.ModelCodes.Gesture)
    # text = translation_model.translate(embedding)
    # print(text)
"""

from sign_language_translator import config, languages, models, text, utils, vision
from sign_language_translator.config import enums
from sign_language_translator.config.assets import Assets
from sign_language_translator.config.enums import ModelCodeGroups, ModelCodes
from sign_language_translator.config.enums import SignFormats as SignFormatCodes
from sign_language_translator.config.enums import SignLanguages as SignLanguageCodes
from sign_language_translator.config.enums import TextLanguages as TextLanguageCodes
from sign_language_translator.config.settings import Settings
from sign_language_translator.languages import get_sign_language, get_text_language
from sign_language_translator.models import get_model
from sign_language_translator.vision._utils import get_sign_wrapper_class
from sign_language_translator.vision.landmarks.landmarks import Landmarks
from sign_language_translator.vision.video.video import Video

__version__ = config.utils.get_package_version()

__all__ = [
    # config
    "Assets",
    "Settings",
    "__version__",
    # modules
    "vision",
    "text",
    "models",
    "languages",
    "utils",
    "config",
    "enums",
    # classes (enum)
    "ModelCodes",
    "TextLanguageCodes",
    "SignLanguageCodes",
    "SignFormatCodes",
    "ModelCodeGroups",
    # classes (wrappers)
    "Landmarks",
    "Video",
    # object loaders / factory functions
    "get_sign_language",
    "get_text_language",
    "get_model",
    "get_sign_wrapper_class",
]



"""
This module defines the `Landmarks` wrapper class which inherits from the `Sign` class.
It is used to represent and manipulate landmarks data extracted from a video featuring 1 person.
It provides methods for saving & loading landmarks data from various sources
(including CSV, NPY, PT, and PTH files, numpy arrays, PyTorch tensors, and nested lists).
It enables data augmentation using the `.transform()` method and smart concatenation across time dimension.
It can visualize the sequence of 3D landmarks using the `.show()` method.

Classes:
    Landmarks: A class to represent and manipulate landmarks data.

Example:

    .. code-block:: python

        from sign_language_translator.vision.landmarks.landmarks import Landmarks

        landmarks = Landmarks([[[0,1,2], [1,2,3]]])  # 1 frames, 2 landmarks, 3 coordinates
        landmarks.show()
        landmarks.save('landmarks_file.csv')
        # landmarks = Landmarks.load('landmarks_file.csv')

        landmarks = Landmarks.load_asset('landmarks/pk-hfad-1_car.landmarks-mediapipe-world.csv')
        print(landmarks.shape)
        # (60, 75, 5)
"""

from __future__ import annotations

__all__ = [
    "Landmarks",
]

import os
from os.path import basename
from string import ascii_letters, digits
from typing import Callable, Iterable, List, Literal, Optional, Sequence, Tuple, Union
from warnings import warn

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.animation import FuncAnimation
from numpy.typing import NDArray
from torch import Tensor
from tqdm.auto import tqdm

from sign_language_translator.config.assets import Assets
from sign_language_translator.config.enums import SignFormats
from sign_language_translator.utils import in_jupyter_notebook, validate_path_exists
from sign_language_translator.utils.arrays import ArrayOps
from sign_language_translator.vision.landmarks.connections import (
    BaseConnections,
    get_connections,
)
from sign_language_translator.vision.landmarks.display import (
    MatPlot3D,
    _reset_counter_in_animation_title,
)
from sign_language_translator.vision.sign.sign import Sign


class Landmarks(Sign):
    """A class to represent and manipulate landmarks data. Inherits from the Sign class.

    Args:
        sign (NDArray | Tensor | str): It can be provided as a path to a file (csv, npy, pt, pth), a NumPy array, a PyTorch tensor, or a sequence of arrays or tensors or numbers (3D: n_frames, n_landmarks, n_features).

    Methods:
        load(path: str, **kwargs): Class method to load landmarks data from a file and return a new Landmarks object.
        save(path: str, overwrite=False, precision=4, **kwargs): Saves the landmarks data to a file.
        name(): Static method which returns the string code of the sign format.
        numpy(*args, **kwargs): Returns the landmarks data as a NumPy array.
        torch(dtype=None, device=None): Returns the landmarks data as a PyTorch tensor.
        tolist(): Returns the landmarks data as a nested list.
        concatenate(objects: Iterable[Landmarks]): Concatenates a sequence of Landmarks objects along the first dimension (time) and returns a new Landmarks object
        transform(transformation: Callable): Applies a transformation function to the landmarks data.
        show(**kwargs): Displays the landmarks data.
        __getitem__(indices): Returns a new Landmarks object with the specified indices.
        __iter__(): Initializes the iteration over the frames of the landmarks data.
        __next__(): Returns the next frame of the landmarks data.

    Attributes:
        data: The landmarks data as a NumPy array or PyTorch tensor depending upon what it was initialized with.
        n_frames: The number of frames or time-steps in the data.
        n_landmarks: The number of landmarks in each frame of the data.
        n_features: The number of features per landmark (same as n_coordinates).
        shape: The shape of the landmarks data array as a tuple of integers.
        ndim: The number of dimensions of the landmarks data array (should be 3).
    """

    def __init__(
        self,
        sign: Union[
            str,
            NDArray,
            Tensor,
            Sequence[NDArray],
            Sequence[Tensor],
            # Sequence[Union[float, int]],  # 1D
            # Sequence[Sequence[Union[float, int]]],  # 2D
            Sequence[Sequence[Sequence[Union[float, int]]]],  # 3D
        ],
        connections: Union[BaseConnections, str, None] = None,
        **kwargs,
    ) -> None:
        """Initializes the Landmarks object.

        Args:
            sign (Union[str, NDArray, Tensor, Sequence]): It can be a string (path to a .csv, .npy or .pt/.pth file), an NDArray, a Tensor, or a sequence of these types or numbers but should be 3D data (n_frames, n_landmarks, n_features) except for the `.csv` which should be 2D (n_frames rows and n_landmarks*n_features columns).
        """
        self._data = None
        self._path = None
        self._caption = None
        self.__idx = 0

        self._animation = None
        self._connections = None

        self.__initialize_from_arguments(sign, connections=connections, **kwargs)

    @staticmethod
    def name() -> str:
        return SignFormats.LANDMARKS.value

    # ========================= #
    #    Iterate / get frame    #
    # ========================= #

    def __iter__(self):
        self.__idx = 0
        return self

    def __next__(self) -> Union[Tensor, NDArray]:
        if 0 <= self.__idx < len(self):
            frame = self.data[self.__idx]
            self.__idx += 1
            return frame

        raise StopIteration

    def __getitem__(self, indices) -> Landmarks:
        indices = self.__normalize_slice_indices(indices)
        return Landmarks(self.data[indices], connections=self._connections)

    # ========================= #
    #    Typecast / get data    #
    # ========================= #

    def numpy(self, *args, **kwargs) -> NDArray:
        """Returns the landmarks data as a numpy array.
        Additional arguments are passed to the numpy.array constructor.

        Returns:
            NDArray: The sign data as a NumPy array.

        Example:

        .. code-block:: python

            import sign_language_translator as slt

            landmarks = slt.Landmarks([[[0,1,2], [1,2,3]]])
            landmarks.numpy()
            # array([[[0, 1, 2], [1, 2, 3]]])
        """
        return np.array(self, *args, **kwargs)

    def __array__(self, *args, **kwargs):
        return np.array(self.data, *args, **kwargs)

    def torch(
        self,
        dtype: Optional[torch.dtype] = None,
        device: Union[torch.device, str, None] = None,
    ) -> Tensor:
        """Returns the landmarks data as a PyTorch tensor.

        Args:
            dtype (torch.dtype, optional): The desired data type of the tensor. Defaults to None.
            device (Union[torch.device, str], optional): The desired device for the tensor. Defaults to None.

        Returns:
            torch.Tensor: The sign data as a PyTorch tensor.
        """
        if isinstance(self.data, torch.Tensor):
            return self.data.to(dtype=dtype, device=device)

        return torch.tensor(self.data, dtype=dtype, device=device)

    def tolist(self) -> List[List[List[Union[float, int]]]]:
        """Returns the landmarks data as a 3D nested list of numbers.

        Returns:
            List[List[List[Union[float, int]]]]: The sign data as a nested list.
        """
        return self.data.tolist()

    # ================= #
    #    Concatenate    #
    # ================= #

    @staticmethod
    def concatenate(objects: Iterable[Landmarks]) -> Landmarks:
        """Concatenates a sequence of Landmarks objects along the time dimension (dim=0) and returns a new Landmarks object.

        Args:
            objects (Iterable[Landmarks]): An iterable of Landmarks objects to concatenate.

        Returns:
            Landmarks: A new Landmarks object containing the data concatenated in time dimension.

        Raises:
            ValueError: If the connections of the Landmarks objects are not the same.
        """
        datas = []
        cons = set()
        for obj in objects:
            cons.add(obj.connections.name() if obj._connections else None)
            datas.append(obj.data)

        if len(datas) == 0:
            raise ValueError("No Landmarks objects provided for concatenation.")
        if len(cons) > 1:
            raise ValueError(f"All Landmarks objects must have same connections {cons}")

        return Landmarks(
            ArrayOps.concatenate(datas, dim=0), connections=obj._connections
        )

    def copy(self) -> Landmarks:
        """Creates a deep copy of the Landmarks object.

        Returns:
            Landmarks: A new Landmarks object with the same data and connections.
        """
        return Landmarks(ArrayOps.copy(self.data), connections=self._connections)

    def __copy__(self) -> Landmarks:
        return self.copy()

    # =============== #
    #    Transform    #
    # =============== #

    def transform(
        self,
        transformation: Union[
            Callable[[NDArray], NDArray],
            Callable[[Tensor], Tensor],
        ],
        inplace=False,
    ) -> Landmarks:
        data = transformation(self.data)  # type: ignore

        if inplace:
            self.data = data
            return self

        return Landmarks(data, connections=self._connections)

    # ========================== #
    #    Display / show frame    #
    # ========================== #

    def show(self, player: Literal["jshtml", "html5"] = "jshtml", **kwargs) -> None:
        """Displays the landmarks data as a 3D animation
        in a Jupyter notebook or as a video in a separate window if run from the terminal.

        Args:
            player (Literal['jshtml', 'html5'], optional): The visualization tool to use for displaying the animation. Defaults to "jshtml".
            **kwargs: Additional keyword arguments to pass to the `new_animation` method. See its docstring for details.
        """
        if self._animation is None or kwargs:
            self._animation = self.new_animation(**kwargs)

        if in_jupyter_notebook():
            from IPython.display import HTML, display  # type: ignore

            html = (
                self.animation.to_html5_video()
                if player == "html5"
                else self.animation.to_jshtml()
            )
            display(HTML(html))
            plt.close(self.animation._fig)  # type: ignore # pylint:disable = protected-access
        else:
            self._animation = self.new_animation(**kwargs)  # todo: show old plt object
            plt.pause(len(self) * self.animation._interval / 1000)  # type: ignore
            # plt.close()

        _reset_counter_in_animation_title(self.animation)

    def show_frames_grid(self, rows=3, columns=5, **kwargs):
        """
        Displays a grid of frames equally spaced in time
        drawn as 3D scatter plots & lines connecting the points.

        Args:
            rows (int): The number of rows in the grid. Default is 3.
            columns (int): The number of columns in the grid. Default is 5.
            **kwargs: Additional keyword arguments to be passed to the `slt.vision.landmarks.MatPlot3D.frames_grid` function.
        """
        fig = MatPlot3D.frames_grid(
            self.numpy(),
            (rows, columns),
            **(self.connections.matplot3d_config if self._connections else {}),
            figure_title=kwargs.pop(
                "figure_title", self._caption or basename(self._path or "")
            ),
            **kwargs,
        )
        plt.show(block=False)
        plt.pause(5)
        plt.close(fig)

    def new_animation(
        self,
        title: Optional[str] = "{frame_number}",
        style: Literal["dark_background", "default"] = "default",
        azimuth: float = 20,
        elevation: float = 15,
        roll: float = 0,
        azimuth_delta: float = 0,
        elevation_delta: float = 0,
        roll_delta: float = 0,
        scatter_size: float = 2,
        figure_scale: Optional[float] = 5,
        interval: Union[float, int] = 37,
        repeat_delay: Union[float, int] = 200,
        blit: bool = True,
    ) -> FuncAnimation:
        """Creates a new 3D animation object of the landmarks.

        Args:
            title (Optional[str]): The title of the animation. Can include the placeholder "{frame_number}" to display the frame number. Defaults to "{frame_number}".
            style (Literal["dark_background", "default"]): The color theme of the animation. Defaults to "default".
            azimuth (float): The azimuth angle (rotation around the vertical axis) of the camera view point. Defaults to 20.
            elevation (float): The elevation angle (amount of rise from the horizontal plane) of the camera view point. Defaults to 15.
            roll (float): The roll angle (rotation around the line of sight) of the camera view point. Defaults to 0.
            azimuth_delta (float): The change in azimuth angle per frame. Defaults to 0.
            elevation_delta (float): The change in elevation angle per frame. Defaults to 0.
            roll_delta (float): The change in roll angle per frame. Defaults to 0.
            scatter_size (float): The size of the scatter points. Defaults to 2.
            figure_scale (Optional[float]): The size of the figure. Defaults to 5.
            interval (Union[float, int]): The interval between frames in milliseconds. Defaults to 37.
            repeat_delay (Union[float, int]): The delay between animation replays in milliseconds. Defaults to 200.
            blit (bool): Whether to use blitting for faster updates (non-changing graphic elements are rendered once into a background image). Defaults to True.

        Returns:
            FuncAnimation: The created animation.
        """
        return MatPlot3D.animate(
            self.numpy(),
            **(self.connections.matplot3d_config if self._connections else {}),
            title=title,
            azimuth=azimuth,
            elevation=elevation,
            roll=roll,
            azimuth_delta=azimuth_delta,
            elevation_delta=elevation_delta,
            roll_delta=roll_delta,
            scatter_size=scatter_size,
            figure_scale=figure_scale,
            style=style,
            interval=interval,
            repeat_delay=repeat_delay,
            blit=blit,
        )

    # ========== #
    #    Save    #
    # ========== #

    def save(self, path: str, overwrite=False, precision=4, **kwargs) -> None:
        """Saves the current object's data to a file.
        Supported formats include `.npy`, `.pt`/`.pth` (which contain 3D data) and `.csv` which flattens each frame and puts it into a separate row.
        CSV files also contain a header with letters representing the coordinate axes and numbers identifying the landmark.

        Args:
            path (str): The file path to save the data to.
            overwrite (bool, optional): Whether to overwrite the file if it already exists. Defaults to False.
            precision (int, optional): The number of decimal places for saving floating-point values in CSV. Defaults to 4.

        Raises:
            FileExistsError: If the file already exists and `overwrite` is False.
            ValueError: If the file format is not supported.
        """
        path = validate_path_exists(path, overwrite=overwrite)

        if (file_format := path.rsplit(".", 1)[-1].lower()) == "csv":
            np.savetxt(
                path,
                self.data.reshape(self.n_frames, -1),
                delimiter=",",
                fmt=f"%.{precision}f",
                header=self.__make_csv_header(self.n_landmarks, self.n_features),
                comments="",
            )
        elif file_format in ("pt", "pth"):
            torch.save(self.torch(), path)
        elif file_format == "npy":
            np.save(path, self.numpy())
        # elif file_format == "npz":
        #     np.savez_compressed(path, **{os.path.basename(path): self.numpy()})
        else:
            raise ValueError(f"Unsupported file format: '{file_format}'")

    def save_animation(
        self, path, overwrite=True, writer: Optional[str] = None, **kwargs
    ) -> None:
        """Save the video animation of the landmarks data to a file.

        Args:
            path (str): The path to save the animation file.
            overwrite (bool, optional): Whether to overwrite the file if it already exists. Defaults to True.
            writer (Optional[str], optional): The name of the matplotlib writer to use for saving the animation. Defaults to None.
            **kwargs: Additional keyword arguments to be passed to the `new_animation` method.
        """

        path = validate_path_exists(path, overwrite=overwrite)

        if self._animation is None or kwargs:
            self._animation = self.new_animation(**kwargs)

        with tqdm(total=len(self), desc="Saving animation", unit="frames") as bar:
            self.animation.save(
                path, progress_callback=(lambda *args: bar.update()), writer=writer
            )

        _reset_counter_in_animation_title(self.animation)
        # plt.close()

    def save_frames_grid(
        self,
        path: str,
        rows: int = 3,
        columns: int = 5,
        overwrite=True,
        **kwargs,
    ) -> None:
        """Save an image file of a grid of 3D visualizations of the landmarks data.

        Args:
            path (str): The path to save the image.
            rows (int, optional): The number of rows in the grid. Defaults to 3.
            columns (int, optional): The number of columns in the grid. Defaults to 5.
            overwrite (bool, optional): Whether to overwrite the file if it already exists. Defaults to True.
            **kwargs: Additional keyword arguments to customize the grid passed to the `slt.vision.landmarks.MatPlot3D.frames_grid` function.
        """
        path = validate_path_exists(path, overwrite=overwrite)

        fig = MatPlot3D.frames_grid(
            self.numpy(),
            (rows, columns),
            **(self.connections.matplot3d_config if self._connections else {}),
            figure_title=kwargs.pop(
                "figure_title", self._caption or basename(self._path or "")
            ),
            **kwargs,
        )
        fig.savefig(path)
        plt.close(fig)

    # ========== #
    #    Load    #
    # ========== #

    @classmethod
    def load(cls, path: str, **kwargs) -> Landmarks:
        """Class method to load landmarks data from a file and return a new Landmarks object.
        The supported file extensions are `.npy` & `.pt` with must contain 3D arrays (n_frames, n_landmarks, n_features)
        and `.csv` which must have `n_frames` rows and `n_landmarks * n_features` columns.

        The header row in `.csv` is optional if the filename contains the name of a supported embedding model (see `load_asset` function for example models).
        The columns in the .csv are expected to be in the format: [<axis-letter><landmark-number>,...] (e.g. x0, y0, z0, x1, y1, z1, ..., xn, yn, zn).
        Possible axis-letters: x, y, z, a-w, aa-zz, ... (only the first 3 are required to be in that order).

        Args:
            path (str): The file path to load the data from.

        Returns:
            Landmarks: A new Landmarks object containing the loaded data.
        """
        return cls(path, **kwargs)

    @classmethod
    def load_asset(
        cls,
        label: str,
        archive_name: Optional[str] = None,
        overwrite=False,
        progress_bar=True,
        leave=True,
        **kwargs,
    ) -> Landmarks:
        """Class method to load a landmarks file from a one-time-auto-downloaded dataset archive
        and return a new Landmarks object.

        Args:
            label (str): The filename of the landmarks asset to load. 'landmarks/' is prepended to the label if it does not start with it. An example is 'landmarks/pk-hfad-1_airport.landmarks-mediapipe.csv') for embedding of a dictionary video. General syntax is `landmarks/country-organization-number_text[_person_camera].landmarks-model.extension`.
            archive_name (Optional[str], optional): The name of the archive which contains the landmarks asset. If None, the archive name is inferred from the label. An example is `datasets/pk-hfad-1.landmarks-mediapipe-csv.zip`. General syntax is `datasets/country-organization-number[_person_camera].landmarks-model-extension.zip`. Defaults to None.
            overwrite (bool, optional): Whether to overwrite the landmarks asset if it is already extracted. Defaults to False.
            progress_bar (bool, optional): Whether to display a progress bar while downloading the archive or extracting the asset. Defaults to True.
            leave (bool, optional): Whether to leave the progress bar after the operation is complete. Defaults to True.
            **kwargs: Additional keyword arguments to be passed to the Landmarks constructor.

        Raises:
            FileNotFoundError: If no landmarks assets are found for the given label.

        Warns:
            UserWarning: If multiple landmarks assets match the given label and the only first asset is used.

        Returns:
            Landmarks: An instance of the Landmarks class representing the dataset video embedding that matched the label.

        Example:

            .. code-block:: python

                import sign_language_translator as slt

                # Load a dictionary video's landmark embedding asset
                landmarks = slt.Landmarks.load_asset("pk-hfad-1_airplane.landmarks-mediapipe.csv")

                # Load a replication video's landmarks from the built-in datasets
                landmarks = slt.Landmarks.load_asset("landmarks/pk-hfad-1_airplane_dm0001_front.landmarks-mediapipe.csv", archive_name="datasets/pk-hfad-1_dm0001_front.landmarks-mediapipe-csv.zip")
        """

        if "/" not in label:
            label = f"landmarks/{label}"

        paths = Assets.extract(
            label,
            archive_name_or_regex=archive_name,
            download_archive=True,
            overwrite=overwrite,
            progress_bar=progress_bar,
            leave=leave,
        )

        if len(paths) == 0:
            raise FileNotFoundError(f"No landmarks assets found for '{label}'")
        if len(paths) > 1:
            warn(f"Multiple landmarks assets matched '{label}'. Using first of:{paths}")

        return cls.load(paths[0], **kwargs)

    def __initialize_from_arguments(
        self,
        sign: Union[
            str,
            NDArray,
            Tensor,
            Sequence[NDArray],
            Sequence[Tensor],
            Sequence[Sequence[Sequence[Union[float, int]]]],
        ],
        **kwargs,
    ):
        """
        Initialize the current landmarks object from the provided arguments.
        The sign parameter can be a string (path to a file), an NDArray, a Tensor, or a sequence of these types.
        """
        if isinstance(sign, str):
            self._from_path(sign)
        elif isinstance(sign, (np.ndarray, Tensor)):
            self._from_data(sign)
        elif isinstance(sign, Sequence):
            if any(isinstance(x, Tensor) for x in sign):
                sign = [(s if isinstance(s, Tensor) else torch.tensor(s)) for s in sign]
                sign = (torch.concatenate if sign[0].ndim == 3 else torch.stack)(sign)
                self._from_data(sign)
            elif isinstance(sign[0], (np.ndarray, float, int, Sequence)):
                self._from_data(np.array(sign))
            else:
                raise ValueError("Unsupported data structure for sign argument.")
        else:
            raise TypeError(
                "Unsupported type for sign argument. provide a path (csv, npy, pt, pth)"
                " or an array of shape (n_frames, n_landmarks, n_coordinates)"
            )

        if kwargs.get("connections") is not None:
            self.connections = kwargs["connections"]

    def _from_path(self, path: str):
        """Read the file specified by the provided path and load all required data into the current object.

        Args:
            path (str): The file path to load the data from.

        Raises:
            ValueError: If the file format is unsupported.
        """
        self._path = os.path.abspath(path)

        # infer connections from filename
        if self._connections is None and (
            inferred := self.__infer_landmarks_model_from_filename(basename(path))
        ):
            self.connections = inferred

        if path.endswith(".npy"):
            data = np.load(path)
        elif path.endswith(".pt"):
            data = torch.load(path)
        elif path.endswith(".csv"):
            data = self.__load_csv(path)
        # elif path.endswith(".npz"):
        # elif path.endswith(".json"):
        else:
            raise ValueError(f"Unsupported file format: {path}")

        self._from_data(data)

    def __load_csv(self, path: str, sep=",") -> NDArray:
        if header := self.__get_csv_header(path, sep=sep):
            columns = header.split(sep)
            n_landmarks = len({value.lstrip(ascii_letters) for value in columns})
            n_features = len({value.rstrip(digits) for value in columns})
        else:
            if self._connections is None:
                # ToDo: parse path for model name (currently assuming that it's already been parsed)
                raise ValueError(
                    "Could not infer the number of landmarks and coordinates "
                    f"from the file name '{basename(path)}' nor the header inside the file."
                )
            n_landmarks = self.connections.n_landmarks
            n_features = self.connections.n_features

        data = np.loadtxt(path, delimiter=sep, skiprows=int(bool(header)))

        return data.reshape(data.shape[0], n_landmarks, n_features)

    def _from_data(self, sign_data: Union[NDArray, Tensor]):
        """Initialize the current landmarks object from a provided 3D array.

        Args:
            sign_data (Union[NDArray, Tensor]): A 3D array of shape (n_frames, n_landmarks, n_features).
        """
        if isinstance(sign_data, np.ndarray):
            if not np.issubdtype(sign_data.dtype, np.number):
                raise ValueError("Expected sign_data to be a numeric sequence.")

        if sign_data.ndim != 3:
            raise ValueError(
                "Expected sign_data to be 3D (n_frames, n_landmarks, n_features)"
                f" but got {sign_data.ndim}D"
            )
        self._data = sign_data

    # ================ #
    #    Properties    #
    # ================ #

    @property
    def data(self) -> Union[NDArray, Tensor]:
        """The landmarks data which is a 3D array or tensor of shape (n_frames, n_landmarks, n_features)."""
        if self._data is None:
            raise ValueError("No data has been loaded yet")
        return self._data

    @data.setter
    def data(self, value: Union[NDArray, Tensor]):
        if value.ndim != 3:
            raise ValueError(
                "Expected value to be 3D (n_frames, n_landmarks, n_features)"
                f" but got {value.ndim}D"
            )
        if self._connections and value.shape[1] != self.connections.n_landmarks:
            raise ValueError(
                f"Expected data to have {self.connections.n_landmarks} landmarks, got {value.shape[1]}"
            )
        self._data = value

    def __len__(self) -> int:
        return self.data.shape[0]

    @property
    def n_frames(self) -> int:
        """The number of frames or time-steps in the landmarks data object."""
        return len(self)

    @property
    def n_landmarks(self) -> int:
        """The number of landmarks in each frame."""
        return self.data.shape[1]

    @property
    def n_features(self) -> int:
        """The number of features (coordinates) for each landmark."""
        return self.data.shape[2]

    @property
    def n_coordinates(self) -> int:
        """The number of axes/coordinates (features) for each landmark."""
        return self.n_features

    @property
    def shape(self) -> Tuple[int, ...]:
        """number of elements in each of the data array's dimensions e.g. (n_frames, n_landmarks, n_features)"""
        return tuple(self.data.shape)

    @property
    def ndim(self) -> int:
        """The number of dimensions of the landmarks data array (should be 3)."""
        return self.data.ndim

    @property
    def connections(self) -> BaseConnections:
        """
        Object defining the order in which landmarks are connected during display
        and other properties depending on the model used to extract the landmarks.

        Raises:
            ValueError: If this property is accessed before landmarks connections have been defined.
        """
        if self._connections is None:
            raise ValueError("No landmarks connections have been defined yet.")
        return self._connections

    @connections.setter
    def connections(self, value: Union[BaseConnections, str]):
        if isinstance(value, str):
            value = get_connections(value)
        elif not isinstance(value, BaseConnections) and value is not None:
            raise TypeError(f"Expected BaseConnections object or string, got {value}")

        # check if connections are compatible with data
        if self._data is not None and value.n_landmarks != self.n_landmarks:
            raise ValueError(
                f"Expected connections to have {self.n_landmarks} landmarks, got {value.n_landmarks}"
            )
        self._connections = value

    @property
    def animation(self) -> FuncAnimation:
        """
        Visualization of the landmarks on a 3D graph.

        Note:
            For interactive display in a Jupyter notebook, use `%matplotlib widget` magic command and then run a cell with `landmarks_obj.animation` on last line.
        """
        if self._animation is None:
            self._animation = self.new_animation()
        return self._animation

    # =========== #
    #    Utils    #
    # =========== #

    def __normalize_slice_indices(self, indices) -> Tuple:
        if not isinstance(indices, tuple):
            indices = (indices,)

        # typecast indices into self.data's compatible type
        indices = tuple(
            (
                ArrayOps.cast(idx, type(self.data), _dtype=int)
                if isinstance(idx, (np.ndarray, Tensor))
                else slice(idx, (idx + 1) or None) if isinstance(idx, int) else idx
            )
            for idx in indices
        )
        return indices

    def __make_csv_header(self, n_landmarks: int, n_features: int):
        axes = list("xyz" + ascii_letters[:23])
        if n_features > 26:
            axes += [
                chr(i) + chr(j)
                for i in range(ord("a"), ord("z") + 1)
                for j in range(ord("a"), ord("z") + 1)
            ]
        header = ",".join(
            f"{axes[c]}{l}" for l in range(n_landmarks) for c in range(n_features)
        )
        return header

    def __get_csv_header(self, path: str, sep=",", encoding="utf-8") -> str:
        with open(path, "r", encoding=encoding) as f:
            line = f.readline().strip()

        for value in line.split(sep):
            try:
                if value:
                    float(value)
            except ValueError:
                return line
        return ""

    def __infer_landmarks_model_from_filename(self, filename: str) -> str:
        # todo: make robust!
        # todo: parse using asset name parser
        model_names = [
            "mediapipe-image",
            "mediapipe-world",
            # "mediapipe-all",
            # "mediapipe",
        ]
        for name in model_names:
            if name in filename:
                return name
        return ""
